"""Point3D data."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike
    from builtins import NotImplementedType
    from ansys.edb.core.utility.value import Value

from ansys.edb.core.utility import conversions


class Point3DData:
    """Represents a point on a 3D coordinate system."""

    def __init__(self, x: ValueLike, y: ValueLike, z: ValueLike):
        """Initialize a 3D point.

        Parameters
        ----------
        x : :term:`ValueLike`
        y : :term:`ValueLike`
        z : :term:`ValueLike`
        """
        self.x, self.y, self.z = x, y, z

    def __eq__(self, other: Point3DData) -> bool:
        """Compare two points by exact coordinates."""
        if isinstance(other, Point3DData):
            return self.x == other.x and self.y == other.y and self.z == other.z
        return False

    def __add__(self, other: Point3DData) -> Point3DData | NotImplementedType:
        """Add two points piecewise."""
        if isinstance(other, Point3DData):
            return Point3DData(self.x + other.x, self.y + other.y, self.z + other.z)
        return NotImplemented

    def __sub__(self, other: Point3DData) -> Point3DData | NotImplementedType:
        """Subtract a point piecewise."""
        if isinstance(other, Point3DData):
            return Point3DData(self.x - other.x, self.y - other.y, self.z - other.z)
        return NotImplemented

    def __mul__(self, other: Point3DData) -> Point3DData | NotImplementedType:
        """Compute a cross product if ``Point3DData`` is provided. Otherwise, perform scalar multiplication."""
        if isinstance(other, Point3DData):
            x = self.y * other.z - self.z * other.y
            y = self.z * other.x - self.x * other.y
            z = self.x * other.y - self.y * other.x
            return Point3DData(x, y, z)
        if isinstance(other, (int, float)):
            return Point3DData(self.x * other, self.y * other, self.z * other)
        return NotImplemented

    def __rmul__(self, other: Point3DData) -> Point3DData | NotImplementedType:
        """Perform scalar multiplication."""
        return self.__mul__(other)

    def __div__(self, other: Point3DData) -> Point3DData | NotImplementedType:
        """Perform scalar division."""
        if isinstance(other, (int, float)):
            return Point3DData(self.x / other, self.y / other, self.z / other)
        return NotImplemented

    def __neg__(self) -> Point3DData:
        """Negate the signs on the point coordinates."""
        return Point3DData(0, 0, 0) - self

    @property
    def x(self) -> Value:
        """:class:`.Value`: X coordinate."""
        return self._x

    @x.setter
    def x(self, x: ValueLike):
        self._x = conversions.to_value(x)

    @property
    def y(self) -> Value:
        """:class:`.Value`: Y coordinate."""
        return self._y

    @y.setter
    def y(self, y: ValueLike):
        self._y = conversions.to_value(y)

    @property
    def z(self) -> Value:
        """:class:`.Value`: Z coordinate."""
        return self._z

    @z.setter
    def z(self, z: ValueLike):
        self._z = conversions.to_value(z)

    @property
    def magnitude(self) -> float:
        """
        :obj:`float`: Magnitude or length of the point.

        This property is read-only.
        """
        return self.magnitude_sqr.sqrt.double

    @property
    def magnitude_sqr(self) -> float:
        """
        :obj:`float`: Magnitude-square of the point.

        This property is read-only.
        """
        return self.x * self.x + self.y * self.y + self.z * self.z

    def equals(self, other, tolerance: float = 1e-9) -> bool:
        """
        Compare the equality of two 3D points within a given tolerance.

        Parameters
        ----------
        other : .Point3DData
        tolerance : float, default: 1e-9
            Tolerance.

        Returns
        -------
        bool
            ``True`` if the 3D points are equal, ``False`` otherwise.
        """
        if isinstance(other, Point3DData):
            return (
                self.x.equals(other.x, tolerance)
                and self.y.equals(other.y, tolerance)
                and self.z.equals(other.z, tolerance)
            )
        return False

    def distance(self, other: Point3DData) -> float:
        """
        Compute the distance from this point to another point.

        Parameters
        ----------
        other : .Point3DData
            Other point

        Returns
        -------
        float
           Distance to the other point.
        """
        if isinstance(other, Point3DData):
            return (self - other).magnitude

    def midpoint(self, other: Point3DData) -> Point3DData:
        """
        Compute the midpoint of this point and another point.

        Parameters
        ----------
        other : .Point3DData
            Other point

        Returns
        -------
        .Point3DData
            Midpoint of the two points.
        """
        if isinstance(other, Point3DData):
            return 0.5 * (self + other)



"""Package definition."""

from ansys.api.edb.v1 import package_def_pb2_grpc

from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase, parser
from ansys.edb.core.inner.messages import (
    edb_obj_message,
    get_product_property_ids_message,
    get_product_property_message,
    int_property_message,
    polygon_data_property_message,
    set_heat_sink_message,
    set_product_property_message,
    string_property_message,
    value_message,
    value_property_message,
)
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.heat_sink import HeatSink, HeatSinkFinOrientation
from ansys.edb.core.utility.value import Value


class PackageDef(ObjBase):
    """Represents a package definition."""

    __stub: package_def_pb2_grpc.PackageDefServiceStub = StubAccessor(StubType.package_def)

    @classmethod
    def create(cls, db, name):
        """Create a package definition in a given database.

        Parameters
        ----------
        db :class:`.Database`
            Database to create the package definition in.
        name : str
            Name of the package definition.

        Returns
        -------
        PackageDef
            Package definition created.
        """
        return PackageDef(cls.__stub.Create(string_property_message(db, name)))

    @classmethod
    def find_by_name(cls, db, name):
        """Find a package definition object by name in a given database.

        Parameters
        ----------
        db :class:`.Database`
            Database to search for the package definition.
        name : str
            Name of the package definition.

        Returns
        -------
        PackageDef
            Package definition found, ``None`` otherwise.
        """
        return PackageDef(cls.__stub.FindByName(string_property_message(db, name)))

    @classmethod
    def find_by_id(cls, db, uid):
        """Find a package definition object by ID in a given database.

        Parameters
        ----------
        db :class:`.Database`
            Database to search for the package definition.
        UID : int
            Unique identifier for the package definition.

        Returns
        -------
        PackageDef
            Package definition found, ``None`` otherwise.
        """
        return PackageDef(cls.__stub.FindByEDBUId(int_property_message(db, uid)))

    @property
    def definition_type(self):
        """:class:`.DefinitionObjType`: Definition type."""
        return DefinitionObjType.PACKAGE_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the package definition."""
        return self.__stub.GetName(edb_obj_message(self)).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(string_property_message(self, value))

    @property
    @parser.to_polygon_data
    def exterior_boundary(self):
        """:class:`.PolygonData`: \
        Exterior boundary of the package definition."""
        return self.__stub.GetExteriorBoundary(edb_obj_message(self))

    @exterior_boundary.setter
    def exterior_boundary(self, boundary):
        self.__stub.SetExteriorBoundary(polygon_data_property_message(self, boundary))

    @property
    def height(self):
        """:class:`.Value`: Height of the package."""
        return Value(self.__stub.GetHeight(edb_obj_message(self)))

    @height.setter
    def height(self, height):
        self.__stub.SetHeight(value_property_message(self, value_message(height)))

    @property
    def operating_power(self):
        """:class:`.Value`: Operating power of the package."""
        return Value(self.__stub.GetOperatingPower(edb_obj_message(self)))

    @operating_power.setter
    def operating_power(self, power):
        self.__stub.SetOperatingPower(value_property_message(self, value_message(power)))

    @property
    def maximum_power(self):
        """:class:`.Value`: Maximum power of the package."""
        return Value(self.__stub.GetMaximumPower(edb_obj_message(self)))

    @maximum_power.setter
    def maximum_power(self, power):
        self.__stub.SetMaximumPower(value_property_message(self, value_message(power)))

    @property
    def thermal_conductivity(self):
        """:class:`.Value`: Thermal conductivity of the package."""
        return Value(self.__stub.GetTherm_Cond(edb_obj_message(self)))

    @thermal_conductivity.setter
    def thermal_conductivity(self, conductivity):
        self.__stub.SetTherm_Cond(value_property_message(self, value_message(conductivity)))

    @property
    def theta_jb(self):
        """:class:`.Value`: Theta JB (junction to board) of the package."""
        return Value(self.__stub.GetTheta_JB(edb_obj_message(self)))

    @theta_jb.setter
    def theta_jb(self, theta):
        self.__stub.SetTheta_JB(value_property_message(self, value_message(theta)))

    @property
    def theta_jc(self):
        """:class:`.Value`: Theta JC (junction to case) of the package."""
        return Value(self.__stub.GetTheta_JC(edb_obj_message(self)))

    @theta_jc.setter
    def theta_jc(self, theta):
        self.__stub.SetTheta_JC(value_property_message(self, value_message(theta)))

    @property
    def heat_sink(self):
        """:class:`.HeatSink`: Heat sink model assigned to the package."""
        heat_sink_paramaters = self.__stub.GetHeatSink(edb_obj_message(self))
        return HeatSink(
            heat_sink_paramaters.thickness,
            heat_sink_paramaters.spacing,
            heat_sink_paramaters.base_height,
            heat_sink_paramaters.height,
            HeatSinkFinOrientation(heat_sink_paramaters.orientation),
        )

    @heat_sink.setter
    def heat_sink(self, heat_sink_value):
        self.__stub.SetHeatSink(set_heat_sink_message(self, heat_sink_value))

    def delete(self):
        """Delete the package definition."""
        self.__stub.Delete(edb_obj_message(self))

    def get_product_property(self, prod_id, attr_it):
        """Get the product property for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : ProductIdType
            Product ID.
        attr_it : int
            Attribute ID.

        Returns
        -------
        str
            Product property for the given product ID and attribute ID.
        """
        return self.__stub.GetProductProperty(
            get_product_property_message(self, prod_id, attr_it)
        ).value

    def set_product_property(self, prod_id, attr_it, prop_value):
        """Set the product property for the given product ID and attribute ID.

        Parameters
        ----------
        prod_id : ProductIdType
            Product ID.
        attr_it : int
            Attribute ID.
        prop_value : str
            New value for the product property.
        """
        self.__stub.SetProductProperty(
            set_product_property_message(self, prod_id, attr_it, prop_value)
        )

    def get_product_property_ids(self, prod_id):
        """Get the list of property IDs for a given property ID.

        Parameters
        ----------
        prod_id : ProductIdType
            Product ID.

        Returns
        -------
        list[int]
            Attribute IDs for the given product ID.
        """
        attr_ids = self.__stub.GetProductPropertyIds(
            get_product_property_ids_message(self, prod_id)
        ).ids
        return [attr_id for attr_id in attr_ids]



"""Net class."""

import ansys.api.edb.v1.net_class_pb2 as nc_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import layout_obj, messages
from ansys.edb.core.session import StubAccessor, StubType


class NetClass(layout_obj.LayoutObj):
    """Represents a net class."""

    __stub = StubAccessor(StubType.netclass)
    layout_obj_type = LayoutObjType.NET_CLASS

    @classmethod
    def create(cls, layout, name):
        """
        Create a net.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the net class in.
        name : str
            Name of the net.

        Returns
        -------
        NetClass
            Net class created.
        """
        return NetClass(
            cls.__stub.Create(nc_pb2.NetClassCreationMessage(layout=layout.msg, name=name))
        )

    @classmethod
    def find_by_name(cls, layout, name):
        """
        Find a net class by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the net class.
        name : str
            Name of the net class.

        Returns
        -------
        NetClass
            Net class found. Check the :obj:`is_null <s.NetClass.is_null>` property
            of the returned net class to see if it exists.
        """
        return NetClass(cls.__stub.FindByName(messages.edb_obj_name_message(layout, name)))

    @property
    def name(self):
        """:obj:`str`: Name of the net class."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, newname):
        self.__stub.SetName(messages.edb_obj_name_message(self.msg, newname))

    @property
    def description(self):
        """:obj:`str` : Description of the net class."""
        return self.__stub.GetDescription(self.msg).value

    @description.setter
    def description(self, newdesc):
        self.__stub.SetDescription(messages.string_property_message(self, newdesc))

    @property
    def is_power_ground(self):
        """:class:`bool`: Flag indicating in the net class belongs to the power/ground \
            :class:`NetClass` class.

        This property is read-only.
        """
        return self.__stub.IsPowerGround(messages.edb_obj_message(self.msg)).value

    @property
    def nets(self):
        """:obj:`list` of :class:`.Net`: List of nets in the net class.

        This property is read-only.
        """
        from ansys.edb.core.net.net import Net

        nets = self.__stub.GetNets(self.msg).items
        return [Net(n) for n in nets]

    def add_net(self, net):
        """
        Add a net to the net class.

        Parameters
        ----------
        net : Net
            Net to add.
        """
        return self.__stub.AddNet(nc_pb2.NetClassEditMessage(netclass=self.msg, net=net.msg))

    def remove_net(self, net):
        """
        Remove a net from the net class.

        Parameters
        ----------
        net : Net
            Net to remove.
        """
        self.__stub.RemoveNet(nc_pb2.NetClassEditMessage(netclass=self.msg, net=net.msg))

    def contains_net(self, net):
        """
        Determine if a net exists in the net class.

        Parameters
        ----------
        net : Net
            Net to search for.

        Returns
        -------
        bool
            ``True`` if the net is in the net class, ``False`` otherwise.
        """
        return self.__stub.ContainsNet(
            nc_pb2.NetClassEditMessage(netclass=self.msg, net=net.msg)
        ).value



"""Cell."""

from enum import Enum

import ansys.api.edb.v1.cell_pb2 as cell_pb2
from ansys.api.edb.v1.cell_pb2_grpc import CellServiceStub
import ansys.api.edb.v1.edb_defs_pb2 as edb_defs_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import ObjBase, messages, variable_server
from ansys.edb.core.layout import layout
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup
from ansys.edb.core.utility.hfss_extent_info import (
    HfssExtentInfo,
    HFSSExtentInfoType,
    OpenRegionType,
)
from ansys.edb.core.utility.temperature_settings import TemperatureSettings
from ansys.edb.core.utility.value import Value


class CellType(Enum):
    """Provides an enum representing the types of cells."""

    CIRCUIT_CELL = edb_defs_pb2.CIRCUIT_CELL
    FOOTPRINT_CELL = edb_defs_pb2.FOOTPRINT_CELL


class DesignMode(Enum):
    """Provides an enum representing design modes."""

    GENERAL = edb_defs_pb2.GENERAL_MODE
    IC = edb_defs_pb2.IC_MODE


# dict representing options of HFSS Extents available via API.
HFSS_EXTENT_ARGS = {
    "use_open_region": "bool",
    "extent_type": "HfssExtentsType",
    "open_region_type": "HfssExtentsOpenRegionType",
    "base_polygon": "EDBObjMessage",
    "dielectric_extent_type": "HfssExtentsType",
    "dielectric_base_polygon": "EDBObjMessage",
    "dielectric": "HfssExtentMessage",
    "honor_user_dielectric": "bool",
    "airbox_truncate_at_ground": "bool",
    "airbox_horizontal": "HfssExtentMessage",
    "airbox_vertical_positive": "HfssExtentMessage",
    "airbox_vertical_negative": "HfssExtentMessage",
    "sync_airbox_vertical_extent": "bool",
    "is_pml_visible": "bool",
    "operating_frequency": "ValueMessage",
    "radiation_level": "ValueMessage",
    "user_xy_data_extent_for_vertical_expansion": "bool",
}


def _return_value(value):
    """Return value as it is."""
    return value


def _translate_hfss_extent(hfss_extent_msg):
    """Convert an ``HfssExtentMessage`` object to a tuple of expected values."""
    return hfss_extent_msg.value, hfss_extent_msg.absolute


def _translate_hfss_extents_enums(msg):
    """Convert ``HfssExtent`` enums to get their values."""
    return msg.value


def primitive_helper(msg):
    """Convert a message to a primitive."""
    return Primitive(msg).cast()


# dictionary describing message type and functions to translate them
_HFSS_EXTENT_MESSAGE_HELPER = {
    "HfssExtentMessage": {
        "msg": messages.hfss_extent_message,
        "val": _translate_hfss_extent,
    },
    "bool": {
        "msg": _return_value,
        "val": bool,
    },
    "ValueMessage": {
        "msg": messages.value_message,
        "val": Value,
    },
    "EDBObjMessage": {
        "msg": messages.edb_obj_message,
        "val": primitive_helper,
    },
    "HfssExtentsType": {
        "msg": _translate_hfss_extents_enums,
        "val": HFSSExtentInfoType,
    },
    "HfssExtentsOpenRegionType": {
        "msg": _translate_hfss_extents_enums,
        "val": OpenRegionType,
    },
}


# takes user-provided arbitrary args and a list of allowed keywords
# return a copy including only the valid args
def sanitize_args(args):
    """Extract valid extent options and convert them into messages."""
    return {
        k: _HFSS_EXTENT_MESSAGE_HELPER[HFSS_EXTENT_ARGS[k]]["msg"](args[k])
        for k in filter(lambda k: k in args, HFSS_EXTENT_ARGS.keys())
    }


def parse_args(msg):
    """Extract extent option values from an HFFSS extent message and add them into a dictionary."""
    res = {}
    for attribute in HFSS_EXTENT_ARGS.keys():
        value = getattr(msg, attribute)
        msg_type = HFSS_EXTENT_ARGS[attribute]
        res[attribute] = _HFSS_EXTENT_MESSAGE_HELPER[msg_type]["val"](value)
    return res


class Cell(ObjBase, variable_server.VariableServer):
    """Represents a cell object."""

    __stub: CellServiceStub = StubAccessor(StubType.cell)
    layout_obj_type = LayoutObjType.CELL

    def __init__(self, msg):
        """Initialize a new cell object.

        Parameters
        ----------
        msg : EDBObjMessage
        """
        ObjBase.__init__(self, msg)
        variable_server.VariableServer.__init__(self, self)

    @classmethod
    def create(cls, db, cell_type, cell_name):
        """Create a cell.

        Parameters
        ----------
        db : :class:`.Database`
            Database to create the cell in.
        cell_type : CellType
            Type of the cell to create.
        cell_name : str
            Name of the cell.

        Returns
        -------
        Cell
            Cell created.
        """
        return Cell(
            cls.__stub.Create(
                cell_pb2.CellCreationMessage(database=db.msg, type=cell_type.value, name=cell_name)
            )
        )

    @property
    def layout(self):
        """:class:`.Layout`: Layout of the cell.

        This property is read-only.
        """
        return layout.Layout(self.__stub.GetLayout(self.msg))

    @property
    def flattened_layout(self):
        """:class:`.Layout`: Flattened layout of the cell.

        This property is read-only.
        """
        return layout.Layout(self.__stub.GetFlattenedLayout(self.msg))

    @classmethod
    def find(cls, database, cell_type, name=None, cell_id=None):
        """Find a cell in a database by either name or ID.

        Parameters
        ----------
        database : :class:`.Database`
            Database to search for the cell.
        cell_type : CellType
            Type of the cell.
        name : str, optional
            Name of the cell. The default is ``None``, in which case a name
            is automatically assigned.
        cell_id : int, optional
            ID of the cell. The default is ``None``, in which case an ID
            is automatically assigned.

        Returns
        -------
        Cell
            Cell that was found, ``None`` otherwise.
        """
        cell = Cell(cls.__stub.Find(messages.cell_find_message(database, cell_type, name, cell_id)))
        return None if cell.is_null else cell

    def delete(self):
        """Delete the cell."""
        self.__stub.Delete(self.msg)

    @property
    def database(self):
        """:class:`.Database`: Owning database of the cell.

        This property is read-only.
        """
        from ansys.edb.core.database import Database

        return Database(self.__stub.GetDatabase(self.msg))

    @property
    def is_footprint(self):
        """:obj:`bool`: Flag indicating if the cell is a footprint.

        This property is read-only.
        """
        return self.__stub.IsFootprint(self.msg).value

    @property
    def is_blackbox(self):
        """:obj:`bool`: Flag indicating if the cell is a blackbox."""
        return self.__stub.IsBlackBox(self.msg).value

    @is_blackbox.setter
    def is_blackbox(self, value):
        self.__stub.SetBlackBox(messages.bool_property_message(self, value))

    @property
    def suppress_pads(self):
        """:obj:`bool`: Flag indicating if the internal non-functional pads are suppressed from rendering/analysis."""
        return self.__stub.GetSuppressPads(self.msg).value

    @suppress_pads.setter
    def suppress_pads(self, value):
        self.__stub.SetSuppressPads(messages.bool_property_message(self, value))

    @property
    def anti_pads_always_on(self):
        """:obj:`bool`: Flag indicating whether antipads are always enabled.

        ``True`` if antipads are always enabled, ``False`` otherwise. If antipads
        are not always enabled, they are triggered only when the via center point falls \
        within fill from another net.
        """
        return self.__stub.GetAntiPadsAlwaysOn(self.msg).value

    @anti_pads_always_on.setter
    def anti_pads_always_on(self, value):
        self.__stub.SetAntiPadsAlwaysOn(messages.bool_property_message(self, value))

    @property
    def anti_pads_option(self):
        """:obj:`int`: Mode for activating antipads.

        - ``0`` for center-point intersection
        - ``1`` for always on
        - ``2`` for pad intersection
        """
        return self.__stub.GetAntiPadsOption(self.msg)

    @anti_pads_option.setter
    def anti_pads_option(self, value):
        self.__stub.SetAntiPadsOption(messages.int_property_message(self, value))

    @property
    def is_symbolic_footprint(self):
        """:obj:`bool`: Flag indicating if the cell is a symbolic footprint.

        This property is read-only.
        """
        return self.__stub.IsSymbolicFootprint(self.msg).value

    @property
    def name(self):
        """:obj:`str`: Name of the cell."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(messages.string_property_message(self, value))

    @property
    def design_mode(self):
        """:obj:`DesignMode`: Design mode of the cell."""
        return DesignMode(self.__stub.GetDesignMode(self.msg).mode)

    @design_mode.setter
    def design_mode(self, value):
        self.__stub.SetDesignMode(messages.design_mode_property_message(self, value))

    @property
    def hfss_extent_info(self):
        """:class:`.HfssExtentInfo`: \
        HFSS extents for the cell."""
        msg = self.__stub.GetHfssExtentInfo(self.msg)
        return HfssExtentInfo(**parse_args(msg))

    def set_hfss_extent_info(self, extents):
        """Set the HFSS extents of this cell.

        Parameters
        ----------
        extents : :class:`.HfssExtentInfo`
        """
        self.__stub.SetHfssExtentInfo(
            cell_pb2.CellSetHfssExtentsMessage(
                cell=self.msg, info=messages.hfss_extent_info_message(extents)
            )
        )

    @property
    def temperature_settings(self):
        """:class:`.TemperatureSettings`: \
        Temperature settings."""
        msg = self.__stub.GetTemperatureSettings(self.msg)
        return TemperatureSettings(
            msg.include_temp_dependence, msg.enable_thermal_feedback, Value(msg.temperature)
        )

    @temperature_settings.setter
    def temperature_settings(self, value):
        self.__stub.SetTemperatureSettings(
            messages.cell_set_temperature_settings_message(self, value)
        )

    def cutout(self, included_nets, clipped_nets, clipping_polygon, clean_clipping=True):
        """Cut out an existing cell into a new cell.

        Parameters
        ----------
        included_nets : list[:class:`.Net`]
            Nets to keep after cutout.
        clipped_nets : list[:class:`.Net`]
            Nets to kept and clip at the boundary after cutout.
        clipping_polygon : :class:`.PolygonData`
            Clipping polygon.
        clean_clipping : bool, optional
            Whether to perform clean clipping. The default is ``True``.

        Returns
        -------
        Cell
            Cell created.
        """
        return Cell(
            self.__stub.CutOut(
                messages.cell_cutout_message(
                    self, included_nets, clipped_nets, clipping_polygon, clean_clipping
                )
            )
        )

    def get_product_property_ids(self, prod_id):
        """Get a list of attribute IDS for a given product ID for the cell.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.

        Returns
        -------
        list[int]
            List of user-defined attribute IDs for properties stored in this object.
        """
        ids = self.__stub.GetProductPropertyIds(
            messages.get_product_property_ids_message(self, prod_id)
        ).ids
        return [prop_id for prop_id in ids]

    def get_product_property(self, prod_id, attr_id):
        """Get the product-specific property of the cell.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.
        attr_id : int
            User-defined ID that identifies the string value stored in the property.

        Returns
        -------
        str
            String stored in the product-specific property.
        """
        return self.__stub.GetProductProperty(
            messages.get_product_property_message(self, prod_id, attr_id)
        )

    def set_product_property(self, prod_id, attr_id, prop_value):
        """Set the product property of the cell for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.
        attr_id : int
            User-defined ID that identifies the string value stored in the property.
        prop_value : str
            New string to store in this property.
        """
        self.__stub.SetProductProperty(
            messages.set_product_property_message(self, prod_id, attr_id, prop_value)
        )

    def delete_simulation_setup(self, name):
        """Delete a simulation setup by name.

        Parameters
        ----------
        name : str
            Name of the setup.
        """
        self.__stub.DeleteSimulationSetup(messages.string_property_message(self, name))

    @property
    def simulation_setups(self):
        """:obj:`list` of \
        [:class:`.SimulationSetup`]: \
        List of all simulation setups of the cell.

        .. seealso:: :obj:`add_simulation_setup`, :obj:`delete_simulation_setup`
        """
        return [
            SimulationSetup(msg).cast() for msg in self.__stub.GetSimulationSetups(self.msg).items
        ]

    def generate_auto_hfss_regions(self):
        """Generate auto HFSS regions.

        This method automatically identifies areas for use as HFSS regions in SIwave simulations.
        """
        self.__stub.GenerateAutoHFSSRegions(self.msg)

    def generate_via_smart_box(self, net_name):
        """Generate a via smart box.

        This method automatically identifies the locations of vias and significant geometry around them.

        Parameters
        ----------
        net_name : str
            Name of the net to crawl in the search for vias.

        Returns
        -------
        list[:class:`.PolygonData`]
            List of boxes, one around each via discovered.
        """
        self.__stub.GenerateViaSmartBox(messages.string_property_message(self, net_name))



"""Via group."""

from ansys.api.edb.v1.via_group_pb2_grpc import ViaGroupServiceStub

from ansys.edb.core.hierarchy.group import Group
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.session import StubAccessor, StubType


class ViaGroup(Group):
    """Represents a via group object."""

    __stub: ViaGroupServiceStub = StubAccessor(StubType.via_group)

    @classmethod
    def create_with_primitives(cls, layout, primitives, is_persistent):
        """Create one or more via groups with primitives.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the via groups in.
        primitives : list[:class:`.Primitive`]
            List of primitives to use to create the via groups.
        is_persistent : bool
            Whether to preserve primitives during via group creation. If ``False``
            primitives are deleted during via group creation.

        Returns
        -------
        list[ViaGroup]
            List of newly created via groups.
        """
        via_groups = cls.__stub.CreateWithPrimitives(
            messages.via_group_create_with_primitives_message(layout, primitives, is_persistent)
        ).items
        return [ViaGroup(vg) for vg in via_groups]

    @classmethod
    def create_with_outline(cls, layout, outline, conductivity_ratio, layer, net=None):
        """Create a via group with a polygon outline.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout that owns the via group.
        outline : list[:class:`Point2D <.PointData>`] or \
            list[:class:`.PolygonData`]
            List of primitives to use to create the via group.
        conductivity_ratio : float
        layer : str or :class:`.Layer`
            Placement layer for the via group.
        net : str or :class:`.Net`, default: None
            Net that the via group is to belong to.

        Returns
        -------
        ViaGroup
            Via group created.
        """
        return ViaGroup(
            cls.__stub.CreateWithOutline(
                messages.via_group_create_with_outline_message(
                    layout, outline, conductivity_ratio, layer, net
                )
            )
        )

    @classmethod
    def find(cls, layout, name):
        """Find a via group by name in the given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the via group.
        name : str
            Name of the via group.

        Returns
        -------
        ViaGroup
            Via group that is found, ``None`` otherwise.
        """
        return ViaGroup(cls.__stub.FindByName(messages.object_name_in_layout_message(layout, name)))

    @property
    @parser.to_polygon_data
    def outline(self):
        """:class:`.PolygonData`: Via group outline.

        This property is read-only.
        """
        return self.__stub.GetOutline(self.msg)

    @property
    def conductor_percentage(self):
        """:obj:`float`: Conductor percentage of the via group.

        This property is read-only.
        """
        return self.__stub.GetConductorPercentage(self.msg).value

    @property
    def persistent(self):
        """:obj:`bool`: Flag indicating if the primitives in the via group are persistent.

        This property is read-only.
        """
        return self.__stub.IsPersistent(self.msg).value



"""Variable server class."""

import ansys.api.edb.v1.variable_server_pb2 as variable_server_msgs

from ansys.edb.core.inner.messages import value_message
from ansys.edb.core.session import get_variable_server_stub
from ansys.edb.core.utility.value import Value


class VariableServer:
    """Provides a class that owns variables.

    A variables can be either a database, cell, or component definition object.
    """

    def __init__(self, variable_owner):
        """Initialize the variable server object.

        Parameters
        ----------
        variable_owner : :class:`.Database`, :class:`.Cell`, :class:`.Layout`
            The :class:`.ObjBase` that owns the variables.
        """
        assert variable_owner.id > 0, "Invalid variable owner ID,"
        self.variable_owner = variable_owner

    @property
    def _variable_owner_msg(self):
        return self.variable_owner.msg

    def add_variable(self, name, value, is_param=False):
        """Add a variable.

        Parameters
        ----------
        name : str
            Variable name.
        value : str, int, float, complex, :class:`.Value`
            Value, which can be any type that can be converted to a :class:`.Value`
            instance.
        is_param : bool, default: False
            Whether the new variable is a parameter. The default is ``False``, which means it is a local variable.

        Notes
        -----
        Variables can be added to the following EDB objects:

        * :class:`.Database` (Variable names must begin with a '$'.)
        * :class:`.ComponentDef`
        * :class:`.Cell`
        * :class:`.Layout` (Adds variable to the corresponding
          :class:`.Cell` instance.)

        Examples
        --------
        Add variables to a cell and create a value that references these variables.

        >>> param = Value(33.1)
        >>> cell.add_variable("blue1", param)
        >>> cell.add_variable("blue2", "25mm")
        >>> vv = cell.create_value("blue1 + blue2")
        >>> print(vv.double)
        33.125
        """
        temp = variable_server_msgs.AddVariableMessage(
            variable_owner=self._variable_owner_msg,
            name=name,
            value=value_message(value),
            isparam=is_param,
        )
        get_variable_server_stub().AddVariable(temp)

    def add_menu_variable(self, name, values, is_param=False, index=0):
        """Add a menu variable.

        Parameters
        ----------
        name : str
            Variable name.
        values : list[str, int, float, complex, :class:`.Value`]
            Each element can be any type that can be converted to a :class:`.Value`
            instance.
        is_param : bool, default: False
            Whether the new variable is a parameter. The default is ``False``, which means it is a local variable.
        index : int, default: 0
            Index of the value that is initially selected.
        """
        list_of_vms = []
        for value in values:
            list_of_vms.append(value_message(value))

        temp = variable_server_msgs.AddMenuVariableMessage(
            variable_owner=self._variable_owner_msg,
            name=name,
            values=list_of_vms,
            isparam=is_param,
            index=index,
        )
        get_variable_server_stub().AddMenuVariable(temp)

    def delete_variable(self, name):
        """Delete a variable.

        Parameters
        ----------
        name : str
            Variable name.
        """
        temp = variable_server_msgs.VariableNameMessage(
            variable_owner=self._variable_owner_msg, name=name
        )
        get_variable_server_stub().DeleteVariable(temp)

    def set_variable_value(self, name, new_value):
        """Set a variable to a new value.

        Parameters
        ----------
        name : str
            Variable name.
        new_value : str, int, float, complex, :class:`.Value`
            New value.
        """
        temp = variable_server_msgs.SetVariableMessage(
            variable_owner=self._variable_owner_msg, name=name, value=value_message(new_value)
        )
        get_variable_server_stub().SetVariableValue(temp)

    def get_variable_value(self, name):
        """Get the value for a given variable.

        Parameters
        ----------
        name : str
            Variable name.

        Returns
        -------
        :class:`.Value`
            Variable value.
        """
        temp = variable_server_msgs.VariableNameMessage(
            variable_owner=self._variable_owner_msg, name=name
        )
        return Value(get_variable_server_stub().GetVariableValue(temp))

    def is_parameter(self, name):
        """Determine if the variable is a parameter.

        Parameters
        ----------
        name : str
            Variable name.

        Returns
        -------
        bool
            ``True`` if the variable is a parameter, ``False`` otherwise.
        """
        temp = variable_server_msgs.VariableNameMessage(
            variable_owner=self._variable_owner_msg, name=name
        )
        return get_variable_server_stub().IsParameter(temp).value

    def get_all_variable_names(self):
        """Get all variable names.

        Returns
        -------
        list[str]
            Names of all variables.
        """
        return get_variable_server_stub().GetAllVariableNames(self._variable_owner_msg).names

    def get_variable_desc(self, name):
        """Get the description of a variable.

        Parameters
        ----------
        name : str
            Variable name.

        Returns
        -------
        str
            Description of the variable.
        """
        temp = variable_server_msgs.VariableNameMessage(
            variable_owner=self._variable_owner_msg, name=name
        )
        return get_variable_server_stub().GetVariableDesc(temp).value

    def set_variable_desc(self, name, desc):
        """Set a variable to have a new description.

        Parameters
        ----------
        name : str
            Variable name.
        desc : str
             New variable description.
        """
        temp = variable_server_msgs.SetDescriptionMessage(
            variable_owner=self._variable_owner_msg, name=name, desc=desc
        )
        get_variable_server_stub().SetVariableDesc(temp)

    def create_value(self, val):
        """Create a value instance.

        This value instance can reference variables on the variable server.

        Parameters
        ----------
        val : str, int, float, complex
            Value, which can be any type that can be converted to a :class:`.Value`
            instance.


        Returns
        -------
        :class:`.Value`

        Notes
        -----
        Creating a value from a :class:`.Database` instance can reference variables
        in the :class:`.Database` instance.

        Creating a value from a :class:`.Cell` instance can reference variables in
        both the :class:`.Database` instance and the
        :class:`.Cell` instance

        Creating a value from a :class:`.ComponentDef`
        instance can reference variables in both the :class:`.Database` instance and
        the :class:`.ComponentDef` instance.
        """
        if isinstance(val, str):
            return Value(val, self)

        return Value(val)



"""Layout Component."""
from ansys.api.edb.v1.layout_component_pb2 import ImportExportLayoutComponentMessage

from ansys.edb.core.hierarchy.cell_instance import CellInstance
from ansys.edb.core.inner.layout_obj import ObjBase
from ansys.edb.core.session import LayoutComponentServiceStub, StubAccessor, StubType


def _import_export_layout_component_msg(layout, output_aedb_comp_path):
    return ImportExportLayoutComponentMessage(
        layout=layout.msg, aedbcomp_path=output_aedb_comp_path
    )


class LayoutComponent(ObjBase):
    """Class representing a layout component."""

    __stub: LayoutComponentServiceStub = StubAccessor(StubType.layout_component)

    @classmethod
    def export_layout_component(cls, layout, output_aedb_comp_path):
        """Export a layout component of the provided cell to the specified output file path.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to be exported.
        output_aedb_comp_path : str
            File path to export the .aedbcomp file to.

        Returns
        -------
        bool
            ``True`` if layout component is successfully exported, ``False`` otherwise.
        """
        return cls.__stub.ExportLayoutComponent(
            _import_export_layout_component_msg(layout, output_aedb_comp_path)
        )

    @classmethod
    def import_layout_component(cls, layout, aedb_comp_path):
        """Import a layout component from the .aedbcomp file at the specified file path into the provided layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout the layout component will be imported into.
        aedb_comp_path : str
            File path of the .aedbcomp file to import the layout component from.

        Returns
        -------
        :class:`.LayoutComponent`
            Imported layout component
        """
        return LayoutComponent(
            cls.__stub.ImportLayoutComponent(
                _import_export_layout_component_msg(layout, aedb_comp_path)
            )
        )

    @property
    def cell_instance(self):
        """:class:`.CellInstance`: The underlying cell instance of the layout component."""
        return CellInstance(self.__stub.GetCellInstance(self.msg))



"""Polygon."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.geometry.polygon_data import PolygonData
    from ansys.edb.core.typing import NetLike, LayerLike

from ansys.api.edb.v1 import polygon_pb2, polygon_pb2_grpc

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType


class Polygon(Primitive):
    """Represents a polygon object."""

    __stub: polygon_pb2_grpc.PolygonServiceStub = StubAccessor(StubType.polygon)

    @classmethod
    def create(
        cls, layout: Layout, layer: LayerLike, net: NetLike | None, polygon_data: PolygonData
    ) -> Polygon:
        """Create a polygon.

        Parameters
        ----------
        layout : .Layout
            Layout to create the polygon in.
        layer : :term:`LayerLike`
            Layer to place the polygon on.
        net : :term:`NetLike` or None
            Net of the polygon.
        polygon_data : .PolygonData
            Outer contour of the polygon.

        Returns
        -------
        Polygon
            Polygon created.
        """
        return Polygon(
            cls.__stub.Create(
                polygon_pb2.PolygonCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    points=messages.polygon_data_message(polygon_data),
                )
            )
        )

    @property
    @parser.to_polygon_data
    def polygon_data(self) -> PolygonData:
        """:class:`.PolygonData`: Outer contour of the polygon."""
        return self.__stub.GetPolygonData(self.msg)

    @polygon_data.setter
    def polygon_data(self, poly: PolygonData):
        self.__stub.SetPolygonData(
            polygon_pb2.SetPolygonDataMessage(
                target=self.msg, poly=messages.polygon_data_message(poly)
            )
        )

    @property
    def can_be_zone_primitive(self) -> bool:
        """:obj:`bool`: Flag indicating if a polygon can be a zone.

        This property is read-only.
        """
        return True



"""Dielectric material definition."""
from __future__ import annotations

from ansys.api.edb.v1 import djordjecvic_sarkar_model_pb2_grpc
from google.protobuf import empty_pb2

from ansys.edb.core import session
from ansys.edb.core.definition.dielectric_material_model import DielectricMaterialModel
from ansys.edb.core.inner import messages


class DjordjecvicSarkarModel(DielectricMaterialModel):
    """Represents a Djordjecvic-Sarkar dielectric material model."""

    __stub: djordjecvic_sarkar_model_pb2_grpc.DjordjecvicSarkarModelServiceStub = (
        session.StubAccessor(session.StubType.djordecvic_sarkar_model)
    )

    @classmethod
    def create(cls) -> DjordjecvicSarkarModel:
        """Create a Djordjecvic Sarkar dielectric material model.

        Returns
        -------
        .DjordjecvicSarkarModel
        """
        return DjordjecvicSarkarModel(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def use_dc_relative_conductivity(self) -> bool:
        """:obj:`bool`: Flag indicating whether the DC relative permittivity nominal value is used."""
        return self.__stub.UseDCRelativePermitivity(self.msg).value

    @use_dc_relative_conductivity.setter
    def use_dc_relative_conductivity(self, enabled: bool):
        self.__stub.SetUseDCRelativePermitivity(messages.bool_property_message(self, enabled))

    @property
    def frequency(self) -> float:
        """:obj:`float`: Frequency value."""
        return self.__stub.GetFrequency(self.msg).value

    @frequency.setter
    def frequency(self, frequency: float):
        self.__stub.SetFrequency(messages.double_property_message(self, frequency))

    @property
    def relative_permittivity_at_frequency(self) -> float:
        """:obj:`float`: Relative permittivity value at the specified :meth:`frequency <.frequency>`."""
        return self.__stub.GetRelativePermitivityAtFrequency(self.msg).value

    @relative_permittivity_at_frequency.setter
    def relative_permittivity_at_frequency(self, frequency: float):
        self.__stub.SetRelativePermitivityAtFrequency(
            messages.double_property_message(self, frequency)
        )

    @property
    def loss_tangent_at_frequency(self) -> float:
        """:obj:`float`: Loss tangent value at the specified :meth:`frequency <.frequency>`."""
        return self.__stub.GetLossTangentAtFrequency(self.msg).value

    @loss_tangent_at_frequency.setter
    def loss_tangent_at_frequency(self, frequency: float):
        self.__stub.SetLossTangentAtFrequency(messages.double_property_message(self, frequency))

    @property
    def dc_relative_permittivity(self) -> float:
        """:obj:`float`: DC relative permittivity value."""
        return self.__stub.GetDCRelativePermitivity(self.msg).value

    @dc_relative_permittivity.setter
    def dc_relative_permittivity(self, permittivity: float):
        self.__stub.SetDCRelativePermitivity(messages.double_property_message(self, permittivity))

    @property
    def dc_conductivity(self) -> float:
        """:obj:`float`: DC conductivity value."""
        return self.__stub.GetDCConductivity(self.msg).value

    @dc_conductivity.setter
    def dc_conductivity(self, conductivity: float):
        self.__stub.SetDCConductivity(messages.double_property_message(self, conductivity))



"""Terminal Instance."""

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import conn_obj, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal


class TerminalInstance(conn_obj.ConnObj):
    """Represents a terminal instance."""

    __stub = StubAccessor(StubType.terminal_instance)
    layout_obj_type = LayoutObjType.TERMINAL_INSTANCE

    @classmethod
    def create(cls, layout, cell_instance, name, net_ref):
        """Create a terminal instance.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the terminal instance in.
        cell_instance : :class:`.CellInstance`
            Name of the cell instance to create the terminal instance on.
        name : :obj:`str`
            Name of the terminal instance.
        net_ref : :class:`.Net` or :obj:`str` or None
            Net reference.

        Returns
        -------
        TerminalInstance
        """
        return TerminalInstance(
            cls.__stub.Create(
                messages.term_inst_creation_message(layout, net_ref, cell_instance, name)
            )
        )

    @property
    def owning_cell_instance(self):
        """:class:`.CellInstance`: Cell instance that owns the terminal."""
        from ansys.edb.core.hierarchy import cell_instance

        return cell_instance.CellInstance(self.__stub.GetOwningCellInstance(self.msg))

    @property
    def definition_terminal(self):
        """:class:`Terminal`: Definition terminal, if any."""
        return Terminal(self.__stub.GetDefinitionTerminal(self.msg)).cast()

    @property
    def definition_terminal_name(self):
        """:obj:`str`: Name of the definition terminal."""
        return self.__stub.GetDefinitionTerminalName(self.msg).value



"""Structure 3D."""

from enum import Enum

import ansys.api.edb.v1.structure_3d_pb2 as structure3d_pb2
from ansys.api.edb.v1.structure_3d_pb2_grpc import Structure3DServiceStub

from ansys.edb.core.hierarchy.group import Group
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class MeshClosure(Enum):
    """Provides an enum representing mesh closure types."""

    OPEN_ENDED = structure3d_pb2.OPEN_ENDED
    ENDS_CLOSED = structure3d_pb2.ENDS_CLOSED
    FILLED_CLOSED = structure3d_pb2.FILLED_CLOSED
    UNDEFINED_CLOSURE = structure3d_pb2.UNDEFINED_CLOSURE


class Structure3D(Group):
    """Represents a 3D structure."""

    __stub: Structure3DServiceStub = StubAccessor(StubType.structure3d)

    @classmethod
    def create(cls, layout, name):
        """Create a 3D structure.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the 3D structure in.
        name : str
            Name of the 3D structure.

        Returns
        -------
        Structure3D
            3D structure created.
        """
        return Structure3D(cls.__stub.Create(messages.object_name_in_layout_message(layout, name)))

    def get_material(self, evaluate):
        """Get the material for the 3D structure.

        Parameters
        ----------
        evaluate : bool
            Whether to resolve any references in the material name.

        Returns
        -------
        str
        """
        return self.__stub.GetMaterial(messages.bool_property_message(self, evaluate)).value

    def set_material(self, mat_name):
        """Set material for the 3D structure.

        Parameters
        ----------
        mat_name : str
            Material name.
        """
        self.__stub.SetMaterial(messages.string_property_message(self, mat_name))

    @property
    def thickness(self):
        """:class:`.Value`: Thickness for the 3D structure."""
        return Value(self.__stub.GetThickness(self.msg))

    @thickness.setter
    def thickness(self, value):
        self.__stub.SetThickness(
            messages.value_property_message(self, messages.value_message(value))
        )

    @property
    def mesh_closure(self):
        """:obj:`MeshClosure`: Mesh closure property for the 3D structure."""
        return MeshClosure(self.__stub.GetMeshClosureProp(self.msg).closure_type)

    @mesh_closure.setter
    def mesh_closure(self, value):
        self.__stub.SetMeshClosureProp(messages.set_closure_message(self, value))



"""Bondwire definition."""

from enum import Enum

from ansys.api.edb.v1 import bondwire_def_pb2_grpc
import ansys.api.edb.v1.bondwire_def_pb2 as pb

from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class BondwireDefType(Enum):
    """Enum representing different types of bondwires."""

    APD_BONDWIRE_DEF = pb.APD_BONDWIRE_DEF
    JEDEC4_BONDWIRE_DEF = pb.JEDEC4_BONDWIRE_DEF
    JEDEC5_BONDWIRE_DEF = pb.JEDEC5_BONDWIRE_DEF


class BondwireDef(ObjBase):
    """Represents a bondwire definition."""

    __stub: bondwire_def_pb2_grpc.BondwireDefServiceStub = StubAccessor(StubType.bondwire_def)

    @property
    def definition_type(self):
        """:class:`.DefinitionObjType`: Object type of the bondwire definition."""
        return DefinitionObjType.BONDWIRE_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the bondwire definition.

        This property is read-only.
        """
        return self.__stub.GetName(self.msg)

    def delete(self):
        """Delete the bondwire definition."""
        self.__stub.Delete(self.msg)

    @staticmethod
    def _bondwire_def_str_message(obj, string):
        """Create a bondwire definition string message.

        Parameters
        ----------
        obj : Union[Database, 'ApdBondwireDef', 'Jedec4BondwireDef', 'Jedec5BondwireDef']
        string : str

        Returns
        -------
        pb.BondwireDefStrMessage
        """
        return pb.BondwireDefStrMessage(object=obj.msg, name=string)


class ApdBondwireDef(BondwireDef):
    """Represents an APD bondwire definition."""

    __stub: bondwire_def_pb2_grpc.ApdBondwireDefServiceStub = StubAccessor(
        StubType.apd_bondwire_def
    )

    @classmethod
    def create(cls, database, name):
        """Create an APD bondwire definition in a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to create the APD bondwire definition in.
        name : str
            Name of the APD bondwire definition.

        Returns
        -------
        ApdBondwireDef
            APD bondwire definition created.
        """
        bw_msg = cls.__stub.Create(BondwireDef._bondwire_def_str_message(database, name))
        return ApdBondwireDef(bw_msg)

    @classmethod
    def load_definitions_from_file(cls, database, name):
        """Load an APD bondwire definition into a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to load the APD bondwire into.
        name : str
            Name of the APD bondwire definition.
        """
        cls.__stub.LoadDefinitionsFromFile(BondwireDef._bondwire_def_str_message(database, name))

    @classmethod
    def find_by_name(cls, database, name):
        """Find an APD bondwire definition by name in a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to search for the APD bondwire definition.
        name : str
            Name of the APD bondwire definition.

        Returns
        -------
        ApdBondwireDef
            APD bondwire definition found.
        """
        return ApdBondwireDef(
            cls.__stub.FindByName(BondwireDef._bondwire_def_str_message(database, name))
        )

    def get_parameters(self):
        """Get the parameters of the APD bondwire definition.

        Returns
        -------
        str
            String block of the bondwire parameters.
        """
        return self.__stub.GetParameters(self.msg)

    def set_parameters(self, name):
        """Set parameters of the APD bondwire definition.

        Parameters
        ----------
        name : str
            String block of the bondwire parameters.
        """
        self.__stub.SetParameters(BondwireDef._bondwire_def_str_message(self, name))

    @property
    def bondwire_type(self):
        """:class:`BondwireDefType`: Type of the APD bondwire.

        This property is read-only.
        """
        return BondwireDefType.APD_BONDWIRE_DEF


class Jedec4BondwireDef(BondwireDef):
    """Represents a JEDEC4 bondwire definition."""

    __stub: bondwire_def_pb2_grpc.Jedec4BondwireDefServiceStub = StubAccessor(
        StubType.jedec4_bondwire_def
    )

    @classmethod
    def create(cls, database, name):
        """Create a JEDEC4 bondwire definition.

        Parameters
        ----------
        database : :class:`.Database`
            Database to create the JEDEC4 bondwire definition in.
        name : str
            Name of the JEDEC4 bondwire definition.

        Returns
        -------
        Jedec4BondwireDef
            JEDEC4 bondwire definition created.
        """
        return Jedec4BondwireDef(
            cls.__stub.Create(BondwireDef._bondwire_def_str_message(database, name))
        )

    @classmethod
    def find_by_name(cls, database, name):
        """Find a JEDEC4 bondwire definition by name in a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to search for the JEDEC4 bondwire definition.
        name : str
            Name of the JEDEC4 bondwire definition.

        Returns
        -------
        Jedec4BondwireDef
            JEDEC4 bondwire definition found.
        """
        return Jedec4BondwireDef(
            cls.__stub.FindByName(BondwireDef._bondwire_def_str_message(database, name))
        )

    def get_parameters(self):
        """Get parameters of the JEDEC4 bondwire definition.

        Returns
        -------
        :class:`.Value`
            Bondwire top-to-die distance.
        """
        return Value(self.__stub.GetParameters(self.msg))

    def set_parameters(self, top_to_die_distance):
        """Set parameters of the JEDEC4 bondwire definition.

        Parameters
        ----------
        top_to_die_distance : :class:`.Value`
            Bondwire top-to-die distance.
        """
        self.__stub.SetParameters(
            pb.Jedec4BondwireDefSetParametersMessage(
                target=self.msg, top_to_die_distance=messages.value_message(top_to_die_distance)
            )
        )

    @property
    def bondwire_type(self):
        """:class:`BondwireDefType`: Type of the JEDEC4 bondwire.

        This property is read-only.
        """
        return BondwireDefType.JEDEC4_BONDWIRE_DEF


class Jedec5BondwireDef(BondwireDef):
    """Represents a JEDEC5 bondwire definition."""

    __stub: bondwire_def_pb2_grpc.Jedec5BondwireDefServiceStub = StubAccessor(
        StubType.jedec5_bondwire_def
    )

    @classmethod
    def create(cls, database, name):
        """Create a JEDEC5 bondwire definition.

        Parameters
        ----------
        database : :class:`.Database`
            Database to create the JEDEC5 bondwire definition in.
        name : str
            Name of the JEDEC5 bondwire definition.

        Returns
        -------
        Jedec5BondwireDef
            JEDEC5 bondwire definition created.
        """
        return Jedec5BondwireDef(
            cls.__stub.Create(BondwireDef._bondwire_def_str_message(database, name))
        )

    @classmethod
    def find_by_name(cls, database, name):
        """Find a JEDEC5 bondwire definition by name in a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to search for the JEDEC5 bondwire definition.
        name : str
            Name of the JEDEC5 bondwire definition.

        Returns
        -------
        Jedec5BondwireDef
            JEDEC5 bondwire definition found.
        """
        return Jedec5BondwireDef(
            cls.__stub.FindByName(BondwireDef._bondwire_def_str_message(database, name))
        )

    def get_parameters(self):
        """Get parameters of the JEDEC5 bondwire definition.

        Returns
        -------
        tuple[:class:`.Value`, :class:`.Value`, :class:`.Value`]
            The tuple is in this format: ``(top_to_die_distance,die_pad_angle,lead_pad_angle)``

            - ``top_to_die_distance``: Bondwire top-to-die distance.
            - ``die_pad_angle``: Bondwire die pad angle.
            - ``lead_pad_angle``: Bondwire lead pad angle.
        """
        get_parameters_msg = self.__stub.GetParameters(self.msg)
        return (
            Value(get_parameters_msg.top_to_die_distance),
            Value(get_parameters_msg.die_pad_angle),
            Value(get_parameters_msg.lead_pad_angle),
        )

    def set_parameters(self, top_to_die_distance, die_pad_angle, lead_pad_angle):
        """Set parameters of the JEDEC5 bondwire definition.

        Parameters
        ----------
        top_to_die_distance : :class:`.Value`
            Bondwire top-to-die distance.
        die_pad_angle : :class:`.Value`
            Bondwire die pad angle.
        lead_pad_angle : :class:`.Value`
            Bondwire lead pad angle.
        """
        self.__stub.SetParameters(
            pb.Jedec5BondwireDefSetParametersMessage(
                target=self.msg,
                parameters=pb.Jedec5BondwireDefParametersMessage(
                    top_to_die_distance=messages.value_message(top_to_die_distance),
                    die_pad_angle=messages.value_message(die_pad_angle),
                    lead_pad_angle=messages.value_message(lead_pad_angle),
                ),
            )
        )

    @property
    def bondwire_type(self):
        """:class:`BondwireDefType`: Type of the JEDEC5 bondwire.

        This property is read-only.
        """
        return BondwireDefType.JEDEC5_BONDWIRE_DEF



"""Dielectric material definition."""
from __future__ import annotations

from typing import Tuple

from ansys.api.edb.v1 import debye_model_pb2_grpc
import ansys.api.edb.v1.debye_model_pb2 as pb
from google.protobuf import empty_pb2

from ansys.edb.core import session
from ansys.edb.core.definition.dielectric_material_model import DielectricMaterialModel
from ansys.edb.core.inner import messages


class DebyeModel(DielectricMaterialModel):
    """Represents a Debye dielectric material model."""

    __stub: debye_model_pb2_grpc.DebyeModelServiceStub = session.StubAccessor(
        session.StubType.debye_model
    )

    @classmethod
    def create(cls) -> DebyeModel:
        """Create a Debye dielectric material model.

        Returns
        -------
        .DebyeModel
        """
        return DebyeModel(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def frequency_range(self) -> Tuple[float, float]:
        """:obj:`tuple` of (:obj:`float`, :obj:`float`): Frequency range (low, high)."""
        range_msg = self.__stub.GetFrequencyRange(messages.edb_obj_message(self))
        return range_msg.low.value, range_msg.high.value

    @frequency_range.setter
    def frequency_range(self, freq: Tuple[float, float]):
        low = freq[0]
        high = freq[1]
        self.__stub.SetFrequencyRange(DebyeModel._set_frequency_range_message(self, low, high))

    @property
    def relative_permittivity_at_high_low_frequency(self) -> Tuple[float, float]:
        """:obj:`tuple` of (:obj:`float`, :obj:`float`): \
        Relative permittivity value at the frequency range minimum and maximum.

        Tuple is of the form ``(rel_perm_at_min_freq, rel_perm_at_max_freq)``.
        """
        range_msg = self.__stub.GetRelativePermitivityAtHighLowFrequency(
            messages.edb_obj_message(self)
        )
        return range_msg.low.value, range_msg.high.value

    @relative_permittivity_at_high_low_frequency.setter
    def relative_permittivity_at_high_low_frequency(self, freq: Tuple[float, float]):
        low = freq[0]
        high = freq[1]
        self.__stub.SetRelativePermitivityAtHighLowFrequency(
            DebyeModel._set_frequency_range_message(self, low, high)
        )

    @property
    def loss_tangent_at_high_low_frequency(self) -> Tuple[float, float]:
        """:obj:`tuple` of (:obj:`float`, :obj:`float`): \
        Dielectric loss tangent value at the frequency range minimum and maximum.

        Tuple is of the form ``(diel_loss_tan_at_min_freq, diel_loss_tan_at_max_freq)``.
        """
        range_msg = self.__stub.GetLossTangentAtHighLowFrequency(messages.edb_obj_message(self))
        return range_msg.low.value, range_msg.high.value

    @loss_tangent_at_high_low_frequency.setter
    def loss_tangent_at_high_low_frequency(self, freq: Tuple[float, float]):
        low = freq[0]
        high = freq[1]
        self.__stub.SetLossTangentAtHighLowFrequency(
            DebyeModel._set_frequency_range_message(self, low, high)
        )

    @property
    def is_relative_permitivity_enabled_at_optical_frequency(self) -> bool:
        """:obj:`bool`: Flag indicating if the relative permittivity at optical frequency is enabled."""
        return self.__stub.IsRelativePermitivityEnabledAtOpticalFrequency(self.msg).value

    @is_relative_permitivity_enabled_at_optical_frequency.setter
    def is_relative_permitivity_enabled_at_optical_frequency(self, enabled: bool):
        self.__stub.SetRelativePermitivityEnabledAtOpticalFrequency(
            messages.bool_property_message(self, enabled)
        )

    @property
    def use_dc_conductivity(self) -> bool:
        """:obj:`bool`: Flag indicating if the DC conductivity nominal value is used."""
        return self.__stub.UseDCConductivity(self.msg).value

    @use_dc_conductivity.setter
    def use_dc_conductivity(self, enabled: bool):
        self.__stub.SetUseDCConductivity(messages.bool_property_message(self, enabled))

    @property
    def relative_permittivity_at_optical_frequency(self) -> float:
        """:obj:`float`: Relative permittivity value at the optical frequency."""
        return self.__stub.GetRelativePermitivityAtOpticalFrequency(self.msg).value

    @relative_permittivity_at_optical_frequency.setter
    def relative_permittivity_at_optical_frequency(self, frequency: float):
        self.__stub.SetRelativePermitivityAtOpticalFrequency(
            messages.double_property_message(self, frequency)
        )

    @property
    def dc_conductivity(self) -> float:
        """:obj:`float`: DC conductivity nominal value."""
        return self.__stub.GetDCConductivity(self.msg).value

    @dc_conductivity.setter
    def dc_conductivity(self, conductivity: float):
        self.__stub.SetDCConductivity(messages.double_property_message(self, conductivity))

    @staticmethod
    def _frequency_range_message(low, high):
        return pb.FrequencyRangeMessage(
            low=messages.double_message(low), high=messages.double_message(high)
        )

    @staticmethod
    def _set_frequency_range_message(target, low, high):
        return pb.SetFrequencyRangeMessage(
            target=target.msg,
            range=DebyeModel._frequency_range_message(low, high),
        )



"""Point data."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike, PointLike
    from ansys.edb.core.utility.value import Value
    from typing import Iterable

from functools import reduce
import math
import operator
import sys

from ansys.api.edb.v1 import point_data_pb2_grpc

from ansys.edb.core import session
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.utility import conversions, value


class PointData:
    """Represents arbitrary (x, y) coordinates that exist on a 2D space."""

    __stub: point_data_pb2_grpc.PointDataServiceStub = session.StubAccessor(
        session.StubType.point_data
    )

    def __init__(self, *data: Iterable[ValueLike]):
        """Initialize point data from a list of coordinates.

        Parameters
        ----------
        data : :obj:`list` of :term:`ValueLike`
        """
        self._x = self._y = self._arc_h = None

        if len(data) == 1:
            # try to expand the argument.
            try:
                iter(data[0])
                data = data[0]
            except TypeError:
                pass

        if len(data) == 1:
            self._x = self._arc_h = conversions.to_value(data[0])
            self._y = sys.float_info.max
        elif len(data) == 2:
            self._x = conversions.to_value(data[0])
            self._y = conversions.to_value(data[1])
            if not self._y.is_parametric and self._y == sys.float_info.max:
                self._arc_h = self._x
        else:
            raise TypeError(
                "`PointData` must receive either one value representing arc height or "
                f"two values representing x and y coordinates. - Received '{data}'"
            )

    def __eq__(self, other: PointData) -> bool:
        """Determine if two objects represent the same coordinates.

        Parameters
        ----------
        other : .PointData

        Returns
        -------
        bool
        """
        return self.equals(other)

    def __len__(self) -> int:
        """Return the number of coordinates present.

        Returns
        -------
        int
        """
        return len(self._matrix_values)

    def __add__(self, other: PointLike) -> PointData:
        """Perform matrix addition of two points.

        Parameters
        ----------
        other : :term:`Point2DLike`

        Returns
        -------
        .PointData
        """
        return self.__class__(self._map_reduce(other, operator.__add__))

    def __sub__(self, other: PointLike) -> PointData:
        """Perform matrix subtraction of two points.

        Parameters
        ----------
        other : :term:`Point2DLike`

        Returns
        -------
        .PointData
        """
        return self.__class__(self._map_reduce(other, operator.__sub__))

    def __str__(self) -> str:
        """Generate unique name for the point object.

        Returns
        -------
        str
        """
        coord = ",".join([str(v) for v in self._matrix_values])
        return f"<{coord}>" if self.is_arc else f"({coord})"

    def equals(self, other, tolerance: float = 0.0) -> bool:
        """Determine if two points are located at the same coordinates.

        Parameters
        ----------
        other : :term:`Point2DLike`
        tolerance : float, default: 0.0

        Returns
        -------
        bool
            ``True`` if the two points are located at the same coordinates,
            ``False`` otherwise.
        """

        def value_equals(a, b):
            return a.equals(b, tolerance)

        try:
            return all(self._map_reduce(other, value_equals))
        except TypeError:
            return False

    @property
    def _matrix_values(self) -> list[Value]:
        """
        :obj:`list` of :class:`.Value`: Coordinates of the point as a list of values.

        This property is read-only.
        """
        return [self.arc_height] if self.is_arc else [self.x, self.y]

    def _map_reduce(self, other, op):
        other = conversions.to_point(other)
        return [reduce(op, values) for values in zip(self._matrix_values, other._matrix_values)]

    @property
    def is_arc(self) -> bool:
        """
        :obj:`bool`: Flag indicating if the point represents an arc.

        This property is read-only.
        """
        return self._arc_h is not None

    @property
    def arc_height(self) -> Value:
        """
        :class:`.Value`: Height of the arc.

        This property is read-only.
        """
        return self._arc_h

    @property
    def x(self) -> Value:
        """
        :class:`.Value`: X coordinate.

        This property is read-only.
        """
        return self._x

    @property
    def y(self) -> Value:
        """
        :class:`.Value`: Y coordinate.

        This property is read-only.
        """
        return self._y

    @property
    def is_parametric(self) -> bool:
        """
        :obj:`bool`: Flag indicating if the point contains parametric values (variable expressions).

        This property is read-only.
        """
        return any(val.is_parametric for val in self._matrix_values)

    def magnitude(self) -> float:
        """Get the magnitude of the point vector.

        Returns
        -------
        float
            Magnitude of the point vector.
        """
        if self.is_arc:
            return 0
        return math.sqrt(sum([v**2 for v in self._matrix_values], value.Value(0)).value)

    def normalized(self) -> PointData:
        """Normalize the point vector.

        Returns
        -------
        PointData
        """
        mag = self.magnitude()
        n = [0] * len(self) if mag == 0 else [v / mag for v in self._matrix_values]
        return self.__class__(n)

    @parser.to_point_data
    def closest(self, start: PointLike, end: PointLike) -> PointData | None:
        """Get the closest point on a line segment from the point.

        Parameters
        ----------
        start : :term:`Point2DLike`
            Start point of the line segment.
        end : :term:`Point2DLike`
            End point of the line segment.

        Returns
        -------
        .PointData or :obj:`None`
            Closet PointData or :obj:`None` if either point is an arc.
        """
        if not self.is_arc:
            return self.__stub.ClosestPoint(messages.point_data_with_line_message(self, start, end))

    def distance(self, start: PointLike, end: PointLike = None) -> float:
        """Compute the shortest distance from the point to a line segment when an end point is given. \
        Otherwise, compute the distance between this point and another point.

        Parameters
        ----------
        start : :term:`Point2DLike`
            Start point of the line segment.
        end : :term:`Point2DLike`, default: None
            End point of the line segment.

        Returns
        -------
        float
        """
        if end is None:
            return (self - start).magnitude()
        else:
            return self.__stub.Distance(
                messages.point_data_with_line_message(self, start, end)
            ).value

    def cross(self, other: PointLike) -> Value | None:
        """Compute the cross product of the point vector with another point vector.

        Parameters
        ----------
        other : :term:`Point2DLike`
            Other point vector.

        Returns
        -------
        .Value or :obj:`None`
            Cross product value or :obj:`None` if either point is an arc.
        """
        other = conversions.to_point(other)
        if not self.is_arc and not other.is_arc:
            return self.x * other.y - self.y * other.x

    def move(self, vector: PointLike) -> PointData | None:
        """Move the point by a vector.

        Parameters
        ----------
        vector : :term:`Point2DLike`
           Vector.

        Returns
        -------
        .PointData or :obj:`None`
           PointData after moving or :obj:`None` if either point is an arc.
        """
        vector = conversions.to_point(vector)
        if not self.is_arc and not vector.is_arc:
            return self + vector

    @parser.to_point_data
    def rotate(self, angle: float, center: PointLike) -> PointData | None:
        """Rotate a point at a given center by a given angle.

        Parameters
        ----------
        angle : float
            Angle in radians.
        center : :term:`Point2DLike`
            Center.

        Returns
        -------
        .PointData or :obj:`None`
            PointData after rotating or :obj:`None` if either point is an arc.
        """
        if not self.is_arc:
            return self.__stub.Rotate(messages.point_data_rotate_message(self, center, angle))

    def dot(self, other: PointLike) -> float:
        """Perform per-component multiplication (dot product) of this point and another point.

        Parameters
        ----------
        other : :term:`Point2DLike`
            Other point.

        Returns
        -------
        float
            Dot product of the two points.
        """
        return sum(self._map_reduce(other, operator.__mul__), value.Value(0)).value

    def angle(self, other: PointLike) -> float:
        """Get the angle between this vector and another vector.

        Parameters
        ----------
        other : :term:`Point2DLike`
            Other vector.

        Returns
        -------
        float
            Angle in radians.
        """
        other = conversions.to_point(other)
        return math.acos(self.dot(other) / (self.magnitude() * other.magnitude()))



"""Layout object."""

import ansys.api.edb.v1.layout_obj_pb2 as layout_obj_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import ObjBase
import ansys.edb.core.inner.messages as messages
from ansys.edb.core.session import LayoutObjServiceStub, StubAccessor, StubType


class LayoutObj(ObjBase):
    """Represents a layout object."""

    __stub: LayoutObjServiceStub = StubAccessor(StubType.layout_obj)
    layout_obj_type = LayoutObjType.INVALID_LAYOUT_OBJ

    @property
    def obj_type(self):
        """:class:`.LayoutObjType`: Layout object type.

        This property is read-only.
        """
        return self.layout_obj_type

    @property
    def layout(self):
        """:class:`.Layout`: Layout owning the object.

        This property is read-only.
        """
        from ansys.edb.core.layout import layout

        return layout.Layout(
            self.__stub.GetLayout(LayoutObj._layout_obj_target_msg(self, self.layout_obj_type))
        )

    def delete(self):
        """Delete the layout object."""
        self.__stub.Delete(LayoutObj._layout_obj_target_msg(self, self.layout_obj_type))

    def get_product_property(self, prod_id, attr_id):
        """Get the product property of the layout object for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.
        attr_id : int
            User-defined ID that identifies the string value stored in the property.

        Returns
        -------
        str
            String stored in the product property.
        """
        return self.__stub.GetProductProperty(
            layout_obj_pb2.GetProductPropertyTypeMessage(
                target=messages.get_product_property_message(self, prod_id, attr_id),
                type=self.layout_obj_type.value,
            )
        ).value

    def set_product_property(self, prod_id, attr_id, prop_value):
        """Set the product property of the layout object for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.
        attr_id : int
           User-defined ID that identifies the string value stored in the property.
        prop_value : str
            String stored in the property.
        """
        self.__stub.SetProductProperty(
            layout_obj_pb2.SetProductPropertyTypeMessage(
                target=messages.set_product_property_message(self, prod_id, attr_id, prop_value),
                type=self.layout_obj_type.value,
            )
        )

    def get_product_property_ids(self, prod_id):
        """Get a list of attribute IDs given a product ID for the layout object.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            ID representing a product that supports the EDB.

        Returns
        -------
        list[int]
            All user-defined attribute IDs for properties stored in the object
        """
        attr_ids = self.__stub.GetProductPropertyIds(
            layout_obj_pb2.GetProductPropertyIdsTypeMessage(
                target=messages.get_product_property_ids_message(self, prod_id),
                type=self.layout_obj_type.value,
            )
        ).ids
        return [attr_id for attr_id in attr_ids]

    @staticmethod
    def _layout_obj_target_msg(layout_obj, layout_type):
        return layout_obj_pb2.LayoutObjTargetMessage(target=layout_obj.msg, type=layout_type.value)



"""Terminals."""

from enum import Enum

import ansys.api.edb.v1.term_pb2 as term_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import TypeField, conn_obj, factory, messages, parser
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.port_post_processing_prop import PortPostProcessingProp
from ansys.edb.core.utility.value import Value


class TerminalType(Enum):
    """Provides an enum representing terminal types."""

    EDGE = term_pb2.EDGE_TERM
    POINT = term_pb2.POINT_TERM
    TERM_INST = term_pb2.TERM_INST_TERM
    PADSTACK_INST = term_pb2.PADSTACK_INST_TERM
    BUNDLE = term_pb2.BUNDLE_TERM
    PIN_GROUP = term_pb2.PIN_GROUP_TERM


class BoundaryType(Enum):
    """Provides an enum representing terminal boundary types."""

    PORT = term_pb2.PORT_BOUNDARY
    PEC = term_pb2.PEC_BOUNDARY
    RLC = term_pb2.RLC_BOUNDARY
    CURRENT_SOURCE = term_pb2.CURRENT_SOURCE
    VOLTAGE_SOURCE = term_pb2.VOLTAGE_SOURCE
    NEXXIM_GROUND = term_pb2.NEXXIM_GROUND
    NEXXIM_PORT = term_pb2.NEXXIM_PORT
    DC_TERMINAL = term_pb2.DC_TERMINAL
    VOLTAGE_PROBE = term_pb2.VOLTAGE_PROBE


class SourceTermToGroundType(Enum):
    """Provides an enum representing source terminal-to-ground types."""

    NO_GROUND = term_pb2.NO_GROUND
    NEGATIVE = term_pb2.NEGATIVE
    POSITIVE = term_pb2.POSITIVE


class HfssPIType(Enum):
    """Provides an enum representing HFSS PI types."""

    DEFAULT = term_pb2.PI_DEFAULT
    COAXIAL_OPEN = term_pb2.PI_COAXIAL_OPEN
    COAXIAL_SHORTENED = term_pb2.PI_COAXIAL_SHORTENED
    GAP = term_pb2.PI_GAP
    LUMPED = term_pb2.PI_LUMPED


class Terminal(conn_obj.ConnObj):
    """Represents a terminal object."""

    __stub = StubAccessor(StubType.terminal)
    type = TypeField(None)
    layout_obj_type = LayoutObjType.TERMINAL

    def cast(self, term_type=None):
        """Cast the terminal object to the correct concrete type, fetching the type if necessary.

        Parameters
        ----------
        term_type : TerminalType

        Returns
        -------
        Terminal
        """
        if self.is_null:
            return

        tt = self.type

        if term_type is not None and tt != term_type:
            return
        return factory.create_terminal(self.msg, tt)

    @classmethod
    def find(cls, layout, name):
        """Find a terminal by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
           Layout to search for the terminal.
        name : .str
            Name of the terminal.

        Returns
        -------
        Terminal
        """
        return Terminal(
            cls.__stub.FindByName(messages.term_find_by_name_message(layout, name))
        ).cast(cls.type)

    @property
    def _params(self):
        return self.__stub.GetParams(self.msg)

    @_params.setter
    def _params(self, values):
        self.__stub.SetParams(messages.term_set_params_message(self, **values))

    @property
    def _type(self):
        """:class:`TerminalType`: Terminal type."""
        return TerminalType(self._params.term_type)

    @property
    def bundle_terminal(self):
        """:class:`BundleTerminal`: Bundle terminal that the terminal belongs to, if any."""
        return factory.create_terminal(self._params.bundle_term, TerminalType.BUNDLE)

    @property
    def reference_terminal(self):
        """:class:`Terminal`: Terminal that the terminal references, if any."""
        return Terminal(self._params.ref_term).cast()

    @reference_terminal.setter
    def reference_terminal(self, value):
        self._params = {"ref_term": value}

    @property
    def reference_layer(self):
        """:class:`.Layer`: Layer that the terminal references, if any, by either layer object or name."""
        return Layer(self._params.ref_layer).cast()

    @reference_layer.setter
    def reference_layer(self, value):
        self._params = {"ref_layer": value}

    @property
    def is_interface_terminal(self):
        """:obj:`bool`: Flag indicating if the terminal is an interface."""
        return self._params.is_interface

    @property
    def is_reference_terminal(self):
        """:obj:`bool`: Flag indicating if the terminal is a reference terminal."""
        return self._params.is_reference

    @property
    def use_reference_from_hierarchy(self):
        """:obj:`bool`: Flag indicating if the terminal can use references from the hierarchy."""
        return self._params.use_ref_from_hierarchy

    @use_reference_from_hierarchy.setter
    def use_reference_from_hierarchy(self, value):
        self._params = {"use_ref_from_hierarchy": value}

    @property
    def is_auto_port(self):
        """:obj:`bool`: Flag indicating if the terminal is an auto port."""
        return self._params.is_auto_port

    @is_auto_port.setter
    def is_auto_port(self, value):
        self._params = {"is_auto_port": value}

    @property
    def is_circuit_port(self):
        """:obj:`bool`: Flag indicating if the terminal is a circuit port."""
        return self._params.is_circuit_port

    @is_circuit_port.setter
    def is_circuit_port(self, value):
        self._params = {"is_circuit_port": value}

    @property
    def name(self):
        """:obj:`str`: Name of the terminal."""
        return self._params.name

    @name.setter
    def name(self, value):
        self._params = {"name": value}

    @property
    def boundary_type(self):
        """:class:`BoundaryType`: Boundary type of the terminal."""
        return BoundaryType(self._params.boundary_type)

    @boundary_type.setter
    def boundary_type(self, value):
        self._params = {"boundary_type": value.value}

    @property
    def impedance(self):
        """:class:`.Value`: Impedance of the terminal."""
        return Value(self._params.impedance)

    @impedance.setter
    def impedance(self, value):
        self._params = {"impedance": value}

    @property
    def source_amplitude(self):
        """:class:`.Value`: Source amplitude of the terminal."""
        return Value(self._params.source_amplitude)

    @source_amplitude.setter
    def source_amplitude(self, value):
        self._params = {"source_amplitude": value}

    @property
    def source_phase(self):
        """:class:`.Value`: Source phase of the terminal."""
        return Value(self._params.source_phase)

    @source_phase.setter
    def source_phase(self, value):
        self._params = {"source_phase": value}

    @property
    def term_to_ground(self):
        """:class:`SourceTermToGroundType`: Source terminal-to-ground type."""
        return SourceTermToGroundType(self._params.term_to_ground)

    @term_to_ground.setter
    def term_to_ground(self, value):
        self._params = {"term_to_ground": value.value}

    @property
    def hfss_pi_type(self):
        """:class:`HfssPIType`: HFSS PI type."""
        return HfssPIType(self._params.hfss_pi_type)

    @hfss_pi_type.setter
    def hfss_pi_type(self, value):
        self._params = {"hfss_pi_type": value.value}

    @property
    def model(self):
        """:obj:`str`: S-parameter model."""
        return self._params.s_param_model

    @model.setter
    def model(self, value):
        self._params = {"s_param_model": value}

    @property
    @parser.to_rlc
    def rlc_boundary_parameters(self):
        """:class:`.Rlc`: RLC boundary parameters."""
        return self._params.rlc

    @rlc_boundary_parameters.setter
    def rlc_boundary_parameters(self, value):
        self._params = {"rlc": value}

    @property
    def port_post_processing_prop(self):
        """:class:`.PortPostProcessingProp`: Port postprocessing properties."""
        msg = self._params.port_post_processing_prop
        return PortPostProcessingProp(
            voltage_magnitude=msg.voltage_magnitude,
            voltage_phase=msg.voltage_phase,
            deembed_length=msg.deembed_length,
            renormalization_impedance=msg.renormalization_impedance,
            do_deembed=msg.do_deembed,
            do_deembed_gap_l=msg.do_deembed_gap_length,
            do_renormalize=msg.do_renormalize,
        )

    @port_post_processing_prop.setter
    def port_post_processing_prop(self, value):
        self._params = {"port_post_processing_prop": value}

    def _product_solvers(self, product_id):
        return self.__stub.GetProductSolvers(
            messages.term_get_product_solver_message(self, product_id)
        )

    def product_solver_option(self, product_id, solver_name):
        """Get the name of the product solver option.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.
        solver_name : :obj:`str`
            Name of the solver.

        Returns
        -------
        :obj:`str`
            Name of the product solver option.
        """
        return next(
            solver.option
            for solver in self._product_solvers(product_id)
            if solver.name == solver_name
        )

    def set_product_solver_option(self, product_id, solver_name, option):
        """Set the product solver option.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.
        solver_name : obj:`str`
            Name of the solver.
        option : obj:`str`
           Name of the product solver option.
        """
        self.__stub.SetProductSolverOptions(
            messages.term_set_solver_option_message(self, product_id, solver_name, option)
        )

    def product_solver_names(self, product_id):
        """Get the list of solver names.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.

        Returns
        -------
        list of str
        """
        return [solver.name for solver in self._product_solvers(product_id)]



"""Class representing parameters for the material property thermal modifier."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike

from ansys.edb.core.utility.value import Value


class BasicQuadraticParams:
    """Represents basic quadratic parameters.

    Attributes
    ----------
    temp_ref_val : :term:`ValueLike`
        Temperature reference value in the quadratic model.
    c1_val : :term:`ValueLike`
        C1 value in the quadratic model.
    c2_val : :term:`ValueLike`
        C2 value in the quadratic model.
    """

    def __init__(
        self,
        temp_ref_val: ValueLike = "22cel",
        c1_val: Value = Value(0),
        c2_val: Value = Value(0),
    ):
        """Initialize a basic quadratic parameters object using given values.

        Parameters
        ----------
        temp_ref_val : :term:`ValueLike`, default: ``"22cel"``
            Temperature reference value in the quadratic model.
        c1_val : :term:`ValueLike`
            C1 value in the quadratic model.
        c2_val : :term:`ValueLike`
            C2 value in the quadratic model.
        """
        self.temp_ref_val = Value(temp_ref_val)
        self.c1_val = c1_val
        self.c2_val = c2_val


class AdvancedQuadraticParams:
    """Represents advanced quadratic parameters.

    Attributes
    ----------
    temp_lower_limit_val : :term:`ValueLike`
        Lower temperature limit where the quadratic model is valid.
    temp_upper_limit_val : :term:`ValueLike`
        Upper temperature limit where the quadratic model is valid.
    auto_calc_constant_thermal_modifier_vals : bool
        Flag indicating if the values for the lower and upper constant thermal modifieries
        are to be automatically calculated.
    lower_constant_thermal_modifier_val : :term:`ValueLike`
        Constant thermal modifier value for temperatures less than the lower constant
        thermal modifier value.
    upper_constant_thermal_modifier_val : :term:`ValueLike`
        Constant thermal modifier value for temperatures greater than the upper
        constant thermal modifier value.
    """

    def __init__(
        self,
        temp_lower_limit_val: ValueLike = "-273.15cel",
        temp_upper_limit_val: ValueLike = "1000cel",
        auto_calc_constant_thermal_modifier_vals: bool = True,
        lower_constant_thermal_modifier_val: ValueLike = 1,
        upper_constant_thermal_modifier_val: ValueLike = 1,
    ):
        """Initialize an advanced quadratic parameters object using given values.

        Parameters
        ----------
        temp_lower_limit_val : :term:`ValueLike`, default: ``"-273.15cel"``
            Lower temperature limit where the quadratic model is valid.
        temp_upper_limit_val : :term:`ValueLike`, default: ``"1000cel"``
            Upper temperature limit where the quadratic model is valid.
        auto_calc_constant_thermal_modifier_vals : bool, default: ``True``
            Flag indicating if the values for the lower and upper constant thermal modifieries
            are to be automatically calculated.
        lower_constant_thermal_modifier_val : :term:`ValueLike`, default: 1
            Constant thermal modifier value for temperatures less than the lower constant
            thermal modifier value.
        upper_constant_thermal_modifier_val : :term:`ValueLike`, default: 1
            Constant thermal modifier value for temperatures greater than the upper
            constant thermal modifier value.
        """
        self.temp_lower_limit_val = Value(temp_lower_limit_val)
        self.temp_upper_limit_val = Value(temp_upper_limit_val)
        self.auto_calc_constant_thermal_modifier_vals = auto_calc_constant_thermal_modifier_vals
        self.lower_constant_thermal_modifier_val = lower_constant_thermal_modifier_val
        self.upper_constant_thermal_modifier_val = upper_constant_thermal_modifier_val



"""Component model definition."""
from enum import Enum

from ansys.api.edb.v1.component_model_pb2 import ComponentModelType as pb_comp_model_Type
from ansys.api.edb.v1.component_model_pb2_grpc import (
    ComponentModelServiceStub,
    DynamicLinkComponentModelServiceStub,
    NPortComponentModelServiceStub,
)
import google.protobuf.wrappers_pb2 as proto_wrappers

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType


class ComponentModelType(Enum):
    """Enum representing component model types."""

    N_PORT = pb_comp_model_Type.N_PORT
    DYNAMIC_LINK = pb_comp_model_Type.DYNAMIC_LINK
    UNKNOWN_COMPONENT_MODEL_TYPE = pb_comp_model_Type.UNKNOWN_MODEL_TYPE


class ComponentModel(ObjBase):
    """Represents a component model."""

    __stub: ComponentModelServiceStub = StubAccessor(StubType.component_model)

    @property
    def reference_file(self):
        """:obj:`str`: Name of the reference file associated with the component model."""
        return self.__stub.GetReferenceFile(self.msg).value

    @reference_file.setter
    def reference_file(self, value):
        self.__stub.SetReferenceFile(messages.string_property_message(self, value))

    @classmethod
    def find_by_name(cls, comp_def, value):
        """Find a component model by name in a given component def.

        Parameters
        ----------
        comp_def : :class:`ComponentDef <ansys.edb.core.definition.ComponentDef>`
            Component def to search for the component model.
        value : str
            Name of the component model.

        Returns
        -------
        ComponentModel
            Component model that is found, ``None`` otherwise.
        """
        return ComponentModel(
            cls.__stub.FindByName(messages.string_property_message(comp_def, value))
        ).cast()

    @classmethod
    def find_by_id(cls, comp_def, value):
        """Find a component model by ID in a given component def.

        Parameters
        ----------
        comp_def : :class:`ComponentDef <ansys.edb.core.definition.ComponentDef>`
            Component def to search for the component model.
        value : int
            ID of the component model.

        Returns
        -------
        ComponentModel
            Component model that is found, ``None`` otherwise.
        """
        return ComponentModel(
            cls.__stub.FindById(messages.int_property_message(comp_def, value))
        ).cast()

    @property
    def name(self) -> str:
        """:obj:`str`: The name of the component model.

        This property is read-only.
        """
        return self.__stub.GetName(self.msg).value

    @property
    def component_model_type(self) -> ComponentModelType:
        """:class:`.ComponentModelType`: The type of the component model.

        This property is read-only.
        """
        return ComponentModelType(self.__stub.GetType(self.msg).type)

    @property
    def component_model_id(self) -> int:
        """:obj:`int`: The id of the component model.

        This property is read-only.
        """
        return self.__stub.GetId(self.msg).value

    def cast(self) -> "ComponentModel":
        """Cast the component model object to the correct concrete type.

        Returns
        -------
        .ComponentModel
        """
        comp_model_type = (
            ComponentModelType.UNKNOWN_COMPONENT_MODEL_TYPE
            if self.is_null
            else self.component_model_type
        )
        if comp_model_type == ComponentModelType.N_PORT:
            return NPortComponentModel(self.msg)
        elif comp_model_type == ComponentModelType.DYNAMIC_LINK:
            return DynamicLinkComponentModel(self.msg)
        return ComponentModel(self.msg)


class NPortComponentModel(ComponentModel):
    """Represents an NPort component model."""

    __stub: NPortComponentModelServiceStub = StubAccessor(StubType.nport_component_model)

    @classmethod
    def create(cls, name):
        """Create an NPort component model.

        Parameters
        ----------
        name : str
            Name of the NPport component model.

        Returns
        -------
        NPortComponentModel
            NPort component model created.

        Notes
        -----
        The component model does not belong to a specific database until it is added to a
        :class:`.ComponentDef` instance.
        """
        return NPortComponentModel(cls.__stub.Create(proto_wrappers.StringValue(value=name)))


class DynamicLinkComponentModel(ComponentModel):
    """Represents a dynamic link component model."""

    __stub: DynamicLinkComponentModelServiceStub = StubAccessor(StubType.dyn_link_component_model)

    @classmethod
    def create(cls, name):
        """Create a dynamic link component model.

        Parameters
        ----------
        name : str
            Name of the dynamic link component model.

        Returns
        -------
        DynamicLinkComponentModel
            Dynamic link component model created.

        Notes
        -----
        The component model does not belong to a specific database until it is added to a
        :class:`.ComponentDef` class.
        """
        return DynamicLinkComponentModel(cls.__stub.Create(proto_wrappers.StringValue(value=name)))

    @property
    def design_name(self):
        """:obj:`str`: Name of the design associated with the dynamic link component model."""
        return self.__stub.GetDesignName(self.msg).value

    @design_name.setter
    def design_name(self, value):
        self.__stub.SetDesignName(messages.string_property_message(self, value))



"""RLC."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike

from ansys.edb.core.utility.value import Value


class Rlc:
    """Represents an RLC.

    Attributes
    ----------
    r : :term:`ValueLike`, default: 0
        Resistance value. This parameter is only used if
        ``r_enabled=True``.
    r_enabled : bool, default: ``False``
        Whether resistance is enabled.
    l : :term:`ValueLike`, default: 0
        Inductance value. This parameter is only used if
        ``c-l_enabled=True``.
    l_enabled : bool, default: ``False``
        Whether inductance is enabled.
    c : :term:`ValueLike`, default: 0
        Capacitance value. This parameter is only used if
        ``c_enabled=True``.
    c_enabled : bool, default: ``False``
        Whether capacitance is enabled.
    is_parallel : bool, default: ``True``
        Whether the r, l, and c elements are in parallel. If ``False``, these
        elements are in series.
    """

    def __init__(
        self,
        r: ValueLike = Value(0),
        r_enabled: bool = False,
        l: ValueLike = Value(0),
        l_enabled: bool = False,
        c: ValueLike = Value(0),
        c_enabled: bool = False,
        is_parallel: bool = True,
    ):
        """Initialize an RLC object using given values.

        Parameters
        ----------
        r : :term:`ValueLike`, default: 0
            Resistance value. This parameter is only used if ``r_enabled=True``.
        r_enabled : bool, default: ``False``
            Whether resistance is enabled.
        l : :term:`ValueLike`, default: 0
            Inductance value. This parameter is only used if ``c-l_enabled=True``.
        l_enabled : bool, default: ``False``
            Whether inductance is enabled.
        c : :term:`ValueLike`, default: 0
            Capacitance value. This parameter is only used if ``c_enabled=True``.
        c_enabled : bool, default: ``False``
            Whether capacitance is enabled.
        is_parallel : bool, default: ``True``
            Whether the r, l, and c elements are in parallel. If ``False``, these
            elements are in series.
        """
        self.r = r
        self.l = l
        self.c = c
        self.r_enabled = r_enabled
        self.l_enabled = l_enabled
        self.c_enabled = c_enabled
        self.is_parallel = is_parallel


class PinPair:
    """Represents a pin pair.

    Attributes
    ----------
    pin1: str
        Name of the first pin.
    pin2: str
        Name of the second pin.
    """

    def __init__(self, pin1, pin2):
        """Initialize a pin pair object.

        Parameters
        ----------
        pin1: str
            Name of the first pin.
        pin2: str
            Name of the second pin.
        """
        self.pin1 = pin1
        self.pin2 = pin2


class PinPairRlc:
    """Represents a pin pair RLC.

    Attributes
    ----------
    pin_pair : .PinPair
        Pin pair property.
    rlc : .Rlc
        RLC value.
    """

    def __init__(self, pin_pair, rlc):
        """Initialize a pin pair RLC object.

        Parameters
        ----------
        pin_pair: .PinPair
            Pin pair property.
        rlc: .Rlc
            RLC property.
        """
        self.pin_pair = pin_pair
        self.rlc = rlc



"""Layout object instance geometry."""

from ansys.api.edb.v1.layout_obj_instance_geometry_pb2 import LayoutObjInstanceGeometryMessage

from ansys.edb.core.inner import ObjBase
from ansys.edb.core.inner.messages import edb_obj_message
from ansys.edb.core.session import LayoutObjInstanceGeometryServiceStub, StubAccessor, StubType


class LayoutObjInstanceGeometry(ObjBase):
    """Represents layout object instance geometry."""

    __stub: LayoutObjInstanceGeometryServiceStub = StubAccessor(
        StubType.layout_obj_instance_geometry
    )

    def __init__(self, geometry, owning_drawing, placement_lyr):
        """Initialize the layout object instance geometry object.

        Parameters
        ----------
        geometry : EDBObjMessage
        owning_drawing : EDBObjMessage
        placement_lyr : EDBObjMessage
        """
        super().__init__(geometry)
        self._owning_drawing_id = owning_drawing.id
        self._placement_layer_id = placement_lyr.id

    @ObjBase.msg.getter
    def msg(self):
        """Protobuf message that represents this object's ID.

        Returns
        -------
        LayoutObjInstanceGeometryMessage
        """
        return LayoutObjInstanceGeometryMessage(
            geometry=super().msg,
            owning_drawing=edb_obj_message(self._owning_drawing_id),
            placement_layer=edb_obj_message(self._placement_layer_id),
        )

    @property
    def material(self):
        """:obj:`str`: Material of the geometry.

        This property is read-only.
        """
        return self.__stub.GetMaterial(self.msg).value

    @property
    def color(self):
        """:obj:`int`: Color of the geometry.

        This property is read-only.
        """
        return self.__stub.GetColor(self.msg).value



"""Simulation Setup."""

from enum import Enum
from typing import List, Union

from ansys.api.edb.v1 import edb_defs_pb2
from ansys.api.edb.v1.simulation_setup_pb2 import (
    HFSSRegionComputeResourceMessage,
    InterpolatingSweepDataMessage,
    SimulationSetupCreationMessage,
    SweepDataListMessage,
    SweepDataListPropertyMessage,
    SweepDataMessage,
)

from ansys.edb.core.inner import messages
from ansys.edb.core.inner.base import ObjBase
from ansys.edb.core.inner.utils import map_list
from ansys.edb.core.session import SimulationSetupServiceStub, StubAccessor, StubType
from ansys.edb.core.simulation_setup.adaptive_solutions import MatrixConvergenceDataEntry


class SimulationSetupType(Enum):
    """Enum representing available simulation setup types."""

    HFSS = edb_defs_pb2.HFSS_SIM
    SI_WAVE = edb_defs_pb2.SI_WAVE_SIM
    SI_WAVE_DCIR = edb_defs_pb2.SI_WAVE_DCIR_SIM
    HFSS_PI = edb_defs_pb2.HFSS_PI_SIM
    RAPTOR_X = edb_defs_pb2.RAPTOR_X_SIM
    Q3D_SIM = edb_defs_pb2.Q3D_SIM


class FreqSweepType(Enum):
    """Enum representing frequency sweep types."""

    INTERPOLATING_SWEEP = edb_defs_pb2.INTERPOLATING_SWEEP
    DISCRETE_SWEEP = edb_defs_pb2.DISCRETE_SWEEP
    BROADBAND_SWEEP = edb_defs_pb2.BROADBAND_SWEEP


class Distribution(Enum):
    """Enum representing frequency distribution types."""

    LIN = "LIN"
    LINC = "LINC"
    ESTP = "ESTP"
    DEC = "DEC"
    OCT = "OCT"


class HFSSRegionComputeResource:
    """Class representing HFSS region computation resources.

    Attributes
    ----------
    region_id : int
      HFSS region id
    resource_percentage : float
      Resource percentage to allocate to HFSS region
    """

    def __init__(self, region_id, resource_percentage):
        """Construct a HFSSRegionComputeResource object using given values.

        Parameters
        ----------
        region_id : int
        resource_percentage : float
        """
        self.region_id = region_id
        self.resource_percentage = resource_percentage


class InterpolatingSweepData:
    """Class representing interpolating frequency sweep data.

    Attributes
    ----------
    relative_s_error : float
      S-parameter error tolerance.
    enforce_causality : bool
      Enforce causality.
    enforce_passivity : bool
      Enforce passivity.
    passivity_tolerance : float
      Tolerance for passivity enforcement.
    max_solutions : int
      Max solutions.
    use_s_matrix : bool
      Use S matrix.
    use_port_impedance : bool
      Use port impedance.
    use_prop_const : bool
      Use propagation constants.
    use_full_basis : bool
      Use full basis.
    fast_sweep : bool
      Use fast sweep.
    adaptive_sampling : bool
      Use adaptive sampling.
    matrix_conv_entry_list : list[MatrixConvergenceDataEntry]
      Matrix convergence data for frequency sweep.
    min_subranges : int
      Minimum number of subranges.
    min_solutions : int
      Minimum number of solutions.
    """

    def __init__(self):
        """Initialize interpolating sweep data settings."""
        self.relative_s_error = 0.005
        self.enforce_causality = False
        self.enforce_passivity = True
        self.passivity_tolerance = 0.0001
        self.max_solutions = 250
        self.use_s_matrix = True
        self.use_port_impedance = True
        self.use_prop_const = True
        self.use_full_basis = True
        self.fast_sweep = False
        self.adaptive_sampling = False
        self.matrix_conv_entry_list = []
        self.min_subranges = 1
        self.min_solutions = 0


class FrequencyData:
    r"""Class representing a frequency setting.

    Attributes
    ----------
    distribution : .Distribution
      Sweep distribution type (see table below).
    start_f : str
      Start frequency is number with optional frequency units.
    end_f : str
      End frequency is number with optional frequency units.
    step : str
      Step is either frequency with optional frequency units or an integer

    Notes
    -----
    Here are the choices for the distribution parameter:

    .. list-table:: Values for distribution parameter
       :widths: 20 45 25
       :header-rows: 1

       * - Distribution
         - Description
         - Example
       * - LIN
         - linear (start, stop, step)
         - LIN 2GHz 4GHz 100MHz or LIN 1dBm 10dBm 1dB
       * - LINC
         - linear (start, stop, count)
         - LINC 2GHz 4GHz 11
       * - ESTP
         - Exponential step (start, stop, count)
         - ESTP 2MHz 10MHz 3
       * - DEC
         - decade (start, stop, number of decades)
         - DEC 10KHz 10GHz 6
       * - OCT
         - octave (start, stop, number of octaves)
         - OCT 10MHz 160MHz 5

    """

    def __init__(self, distribution: Distribution, start_f: str, end_f: str, step: str):
        """Initialize a frequency setting."""
        self.distribution = distribution
        self.start_f = start_f
        self.end_f = end_f
        self.step = step


class SweepData:
    r"""Class representing a sweep data setting.

    Attributes
    ----------
    name : str
      Name of this sweep.
    enabled : bool
      True if this is enabled.
    type : FreqSweepType
      Type of sweep.
    use_q3d_for_dc : bool
      Use Q3D solver for dc calculations.
    save_fields : bool
      Save fields during simulations.
    save_rad_fields_only : bool
      Save only radiation fields during simulations.
    compute_dc_point : bool
      Calculate DC point during simulations.
    siwave_with_3dddm : bool
      SIwave with 3D DDM.
    use_hfss_solver_regions : bool
      SIwave with HFSS solver regions.
    use_hfss_solver_region_sch_gen : bool
      SIwave with HFSS solver regions schematic.
    use_hfss_solver_region_parallel_solve : bool
      SIwave with HFSS solver parallel region simulation.
    use_adp_solution_for_all_sweep_freq : bool
      Enable Using Adaptive Solution for all frequency points.
    num_parallel_hfss_regions : int
      SIwave with HFSS regions: number of regions to solve in parallel.
    parallel_hfss_regions_sim_cfg : list[HFSSRegionComputeResource]
      List of CHFSSRegionComputeResource assignments for parallel region sims.
    auto_s_mat_only_solve : bool
      Auto/Manual SMatrix only solve.
    min_freq_s_mat_only_solve : str
      Minimum frequency SMatrix only solve.
    steady_state_start : float
      Frequency of Steady State Start.
    mesh_freq_choice : int
      Meshing frequencies of the observation mesh.
    mesh_freq_points : list[str]
      Frequency points in meshing frequencies.
    mesh_freq_range_start : str
      Start meshing frequency.
    mesh_freq_range_stop : str
      Stop meshing frequency.
    interpolation_data : InterpolatingSweepData
      Data for interpolating frequency sweeps.
    frequency_data : .FrequencyData or list of .FrequencyData
      List of frequency data.

    """

    def __init__(self, name: str, frequency_data: Union[FrequencyData, List[FrequencyData]]):
        """Initialize a sweep data setting."""
        self.name = name
        self.enabled = True
        self.type = FreqSweepType.INTERPOLATING_SWEEP
        self.use_q3d_for_dc = False
        self.save_fields = False
        self.save_rad_fields_only = False
        self.compute_dc_point = False
        self.siwave_with_3dddm = False
        self.use_hfss_solver_regions = False
        self.use_hfss_solver_region_sch_gen = False
        self.use_hfss_solver_region_parallel_solve = False
        self.use_adp_solution_for_all_sweep_freq = False
        self.num_parallel_hfss_regions = 1
        self.parallel_hfss_regions_sim_cfg = []
        self.auto_s_mat_only_solve = True
        self.min_freq_s_mat_only_solve = "1MHz"
        self.steady_state_start = -1.0
        self.mesh_freq_choice = -1
        self.mesh_freq_points = []
        self.mesh_freq_range_start = "-1.0"
        self.mesh_freq_range_stop = "-1.0"
        self.interpolation_data = InterpolatingSweepData()
        self.frequency_data = frequency_data

    @property
    def frequency_string(self):
        """:obj:`str`: String representing the frequency sweep data.

        This property is read-only.
        """
        frequencies = []
        if isinstance(self.frequency_data, FrequencyData):
            frequencies.append(self.frequency_data)
        elif isinstance(self.frequency_data, list):
            frequencies = self.frequency_data

        freq_str = ""
        for freq in frequencies:
            freq_str += f"{freq.distribution.value} {freq.start_f} {freq.end_f} {freq.step}\t\n"
        return freq_str


def _hfss_region_compute_resource_message(res):
    """Create a ``HFSSRegionComputeResourceMessage`` from a ``HFSSRegionComputeResource``."""
    return HFSSRegionComputeResourceMessage(
        region_id=res.region_id, compute_resource_percentage=res.resource_percentage
    )


def _interpolating_sweep_data_msg(interp_sweep_data):
    """Create a ``InterpolatingSweepDataMessage`` from a ``InterpolatingSweepData``."""
    return InterpolatingSweepDataMessage(
        relative_s_error=interp_sweep_data.relative_s_error,
        enforce_causality=interp_sweep_data.enforce_causality,
        enforce_passivity=interp_sweep_data.enforce_passivity,
        passivity_tolerance=interp_sweep_data.passivity_tolerance,
        max_solutions=interp_sweep_data.max_solutions,
        interp_use_s_matrix=interp_sweep_data.use_s_matrix,
        interp_use_port_impedance=interp_sweep_data.use_port_impedance,
        interp_use_prop_const=interp_sweep_data.use_prop_const,
        interp_use_full_basis=interp_sweep_data.use_full_basis,
        fast_sweep=interp_sweep_data.fast_sweep,
        adaptive_sampling=interp_sweep_data.adaptive_sampling,
        matrix_conv_entry_list=messages.mx_convergence_entry_msg_list(
            interp_sweep_data.matrix_conv_entry_list
        ),
        interp_min_solutions=interp_sweep_data.min_solutions,
        interp_min_subranges=interp_sweep_data.min_subranges,
    )


def _sweep_data_msg(sweep_data):
    """Create a ``SweepDataMessage`` from a ``SweepData``."""
    return SweepDataMessage(
        name=sweep_data.name,
        frequency_string=sweep_data.frequency_string,
        enabled=sweep_data.enabled,
        type=sweep_data.type.value,
        use_q3d_for_dc=sweep_data.use_q3d_for_dc,
        save_fields=sweep_data.save_fields,
        save_rad_fields_only=sweep_data.save_rad_fields_only,
        compute_dc_point=sweep_data.compute_dc_point,
        siwave_with_3dddm=sweep_data.siwave_with_3dddm,
        use_hfss_solver_regions=sweep_data.use_hfss_solver_regions,
        use_hfss_solver_region_sch_gen=sweep_data.use_hfss_solver_region_sch_gen,
        use_hfss_solver_region_parallel_solve=sweep_data.use_hfss_solver_region_parallel_solve,
        use_adp_solution_for_all_sweep_freq=sweep_data.use_adp_solution_for_all_sweep_freq,
        num_parallel_hfss_regions=sweep_data.num_parallel_hfss_regions,
        parallel_hfss_regions_sim_cfg=map_list(
            sweep_data.parallel_hfss_regions_sim_cfg, _hfss_region_compute_resource_message
        ),
        auto_s_mat_only_solve=sweep_data.auto_s_mat_only_solve,
        min_freq_s_mat_only_solve=sweep_data.min_freq_s_mat_only_solve,
        steady_state_start=sweep_data.steady_state_start,
        mesh_freq_choice=sweep_data.mesh_freq_choice,
        mesh_freq_points=sweep_data.mesh_freq_points,
        mesh_freq_range_start=sweep_data.mesh_freq_range_start,
        mesh_freq_range_stop=sweep_data.mesh_freq_range_stop,
        interp_sweep_data=_interpolating_sweep_data_msg(sweep_data.interpolation_data),
    )


def _msg_to_matrix_convergence_entry(msg):
    """Create a ``MatrixConvergenceDataEntry`` from a ``MatrixConvergenceEntryMessage``."""
    return MatrixConvergenceDataEntry(msg.port_1, msg.port_2, msg.mag_limit, msg.phase_limit)


def _msg_to_hfss_region_compute_resource(msg):
    """Create a ``HFSSRegionComputeResource`` from a ``HFSSRegionComputeResourceMessage``."""
    return HFSSRegionComputeResource(msg.region_id, msg.compute_resource_percentage)


def _msg_to_interpolating_sweep_data(msg):
    """Create a ``InterpolatingSweepData`` from a ``InterpolatingSweepDataMessage``."""
    interp_sweep_data = InterpolatingSweepData()
    interp_sweep_data.relative_s_error = msg.relative_s_error
    interp_sweep_data.enforce_passivity = msg.enforce_passivity
    interp_sweep_data.enforce_causality = msg.enforce_causality
    interp_sweep_data.passivity_tolerance = msg.passivity_tolerance
    interp_sweep_data.max_solutions = msg.max_solutions
    interp_sweep_data.use_s_matrix = msg.interp_use_s_matrix
    interp_sweep_data.use_port_impedance = msg.interp_use_port_impedance
    interp_sweep_data.use_prop_const = msg.interp_use_prop_const
    interp_sweep_data.use_full_basis = msg.interp_use_full_basis
    interp_sweep_data.fast_sweep = msg.fast_sweep
    interp_sweep_data.adaptive_sampling = msg.adaptive_sampling
    interp_sweep_data.matrix_conv_entry_list = map_list(
        msg.matrix_conv_entry_list, _msg_to_matrix_convergence_entry
    )
    interp_sweep_data.min_subranges = msg.interp_min_subranges
    interp_sweep_data.min_solutions = msg.interp_min_solutions
    return interp_sweep_data


def _msg_to_sweep_data(msg):
    """Create a ``SweepData`` from a ``SweepDataMessage``."""
    freq_str_params = msg.frequency_string.split()
    sweep_data = SweepData(
        msg.name,
        FrequencyData(
            Distribution[freq_str_params[0]],
            freq_str_params[1],
            freq_str_params[2],
            freq_str_params[3],
        ),
    )
    sweep_data.enabled = msg.enabled
    sweep_data.type = FreqSweepType(msg.type)
    sweep_data.use_q3d_for_dc = msg.use_q3d_for_dc
    sweep_data.save_fields = msg.save_fields
    sweep_data.save_rad_fields_only = msg.save_rad_fields_only
    sweep_data.compute_dc_point = msg.compute_dc_point
    sweep_data.siwave_with_3dddm = msg.siwave_with_3dddm
    sweep_data.use_hfss_solver_regions = msg.use_hfss_solver_regions
    sweep_data.use_hfss_solver_region_sch_gen = msg.use_hfss_solver_region_sch_gen
    sweep_data.use_hfss_solver_region_parallel_solve = msg.use_hfss_solver_region_parallel_solve
    sweep_data.use_adp_solution_for_all_sweep_freq = msg.use_adp_solution_for_all_sweep_freq
    sweep_data.num_parallel_hfss_regions = msg.num_parallel_hfss_regions
    sweep_data.parallel_hfss_regions_sim_cfg = map_list(
        msg.parallel_hfss_regions_sim_cfg, _msg_to_hfss_region_compute_resource
    )
    sweep_data.auto_s_mat_only_solve = msg.auto_s_mat_only_solve
    sweep_data.min_freq_s_mat_only_solve = msg.min_freq_s_mat_only_solve
    sweep_data.steady_state_start = msg.steady_state_start
    sweep_data.mesh_freq_choice = msg.mesh_freq_choice
    sweep_data.mesh_freq_points = msg.mesh_freq_points
    sweep_data.mesh_freq_range_start = msg.mesh_freq_range_start
    sweep_data.mesh_freq_range_stop = msg.mesh_freq_range_stop
    sweep_data.interpolation_data = _msg_to_interpolating_sweep_data(msg.interp_sweep_data)
    return sweep_data


class SimulationSetup(ObjBase):
    """Class representing base simulation setup data."""

    __stub: SimulationSetupServiceStub = StubAccessor(StubType.sim_setup)

    @classmethod
    def _create(cls, cell, sim_setup_name, sim_setup_type):
        return cls(
            SimulationSetup.__stub.Create(
                SimulationSetupCreationMessage(
                    cell=cell.msg, name=sim_setup_name, type=sim_setup_type.value
                )
            )
        )

    @property
    def name(self):
        """:obj:`str`: Name of simulation setup."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, name):
        self.__stub.SetName(messages.string_property_message(self, name))

    @property
    def position(self):
        """:obj:`int`: Position of the simulation in the setup order."""
        return self.__stub.GetPosition(self.msg).value

    @position.setter
    def position(self, position):
        self.__stub.SetPosition(messages.int_property_message(self, position))

    @property
    def sweep_data(self):
        """:obj:`list` of :class:`.SweepData`: Frequency sweeps of the simulation setup."""
        sweep_data = []
        for sweep_data_msg in self.__stub.GetSweepData(self.msg).sweep_data:
            sweep_data.append(_msg_to_sweep_data(sweep_data_msg))
        return sweep_data

    @sweep_data.setter
    def sweep_data(self, sweep_data):
        sweep_data_msgs = []
        for sweep in sweep_data:
            sweep_data_msgs.append(_sweep_data_msg(sweep))
        self.__stub.SetSweepData(
            SweepDataListPropertyMessage(
                target=self.msg, sweeps=SweepDataListMessage(sweep_data=sweep_data_msgs)
            )
        )

    @property
    def type(self):
        """:class:`.SimulationSetupType`: Type of the simulation setup."""
        return SimulationSetupType(self.__stub.GetType(self.msg).type)

    def cast(self):
        """Cast the base SimulationSetup object to correct subclass, if possible.

        Returns
        -------
        SimulationSetup
        """
        from ansys.edb.core.simulation_setup.hfss_simulation_setup import HfssSimulationSetup
        from ansys.edb.core.simulation_setup.raptor_x_simulation_setup import RaptorXSimulationSetup
        from ansys.edb.core.simulation_setup.siwave_dcir_simulation_setup import (
            SIWaveDCIRSimulationSetup,
        )
        from ansys.edb.core.simulation_setup.siwave_simulation_setup import SIWaveSimulationSetup

        if self.is_null:
            return

        sim_type = self.type
        if sim_type == SimulationSetupType.HFSS:
            return HfssSimulationSetup(self.msg)
        elif sim_type == SimulationSetupType.SI_WAVE:
            return SIWaveSimulationSetup(self.msg)
        elif sim_type == SimulationSetupType.SI_WAVE_DCIR:
            return SIWaveDCIRSimulationSetup(self.msg)
        elif sim_type == SimulationSetupType.RAPTOR_X:
            return RaptorXSimulationSetup(self.msg)



"""PadstackInstance Terminal."""

from ansys.edb.core.inner import TypeField, messages
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal, TerminalType


class PadstackInstanceTerminal(Terminal):
    """Represents a padstack instance terminal."""

    __stub = StubAccessor(StubType.padstack_instance_terminal)
    type = TypeField(TerminalType.PADSTACK_INST)

    @classmethod
    def create(cls, layout, name, padstack_instance, layer, net, is_ref=False):
        """Create a padstack instance terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the padstack instance terminal in.
        name : :obj:`str`
            Name of the padstack instance terminal.
        padstack_instance : :class:`.PadstackInstance`
            Padstack instance.
        layer : :class:`.Layer` or :obj:`str`
            Layer to place the padstack instance terminal on.
        net : :class:`.Net` or :obj:`str` or None
            Net.
        is_ref : :obj:`bool`, default: False
            Whether the padstack instance terminal is a reference terminal.

        Returns
        -------
        PadstackInstanceTerminal
        """
        return PadstackInstanceTerminal(
            cls.__stub.Create(
                messages.padstack_inst_term_creation_message(
                    layout, name, padstack_instance, layer, net, is_ref
                )
            )
        )

    @property
    def params(self):
        """:obj:`list` of :class:`.PadstackInstance` and :class:`.Layer`: Padstack instance and layer."""
        from ansys.edb.core.primitive import padstack_instance

        res = self.__stub.GetParameters(self.msg)
        padstack_instance = padstack_instance.PadstackInstance(res.padstack_instance)
        layer = Layer(res.layer.id).cast()
        return padstack_instance, layer

    @params.setter
    def params(self, params):
        (padstack_instance, layer) = params
        self.__stub.SetParameters(
            messages.padstack_inst_term_set_params_message(self, padstack_instance, layer)
        )

    @property
    def padstack_instance(self):
        """:class:`.PadstackInstance`: Padstack instance of the terminal."""
        return self.params[0]

    @property
    def layer(self):
        """:class:`.Layer`: Layer the terminal is placed on."""
        return self.params[1]



"""RTree class."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.geometry.polygon_data import PolygonData
    from ansys.edb.core.geometry.point_data import PointData

from ansys.api.edb.v1 import r_tree_pb2_grpc
import ansys.api.edb.v1.r_tree_pb2 as pb

from ansys.edb.core.inner import ObjBase, messages, parser
from ansys.edb.core.session import StubAccessor, StubType


class RTreeObj:
    """
    Represents an RTree object.

    Attributes
    ----------
    polygon : .PolygonData
        Polygon representation for the object in the spatial index.
    obj : :class:`.ObjBase`
        Object to store in the index.
    """

    def __init__(
        self,
        polygon: PolygonData,
        obj: ObjBase,
    ):
        """Construct an RTree object using given values.

        Parameters
        ----------
        polygon : .PolygonData
            Polygon representation for the object in the spatial index.
        obj : :class:`.ObjBase`
            Object to store in the index.
        """
        self._unique_id = None
        self.polygon = polygon
        self.obj = obj


class RTree(ObjBase):
    """Provides the base RTree class."""

    __stub: r_tree_pb2_grpc.RTreeServiceStub = StubAccessor(StubType.r_tree)

    def __init__(self, msg):
        """Initialize an RTree object."""
        super().__init__(msg)
        self._id_to_obj = {}
        self._obj_to_id = {}
        self._unique_id = None

    @classmethod
    def create(cls, tolerance: float = 1e-9) -> RTree:
        """Create an RTree.

        Parameters
        ----------
        tolerance : float, default: 1e-9
            Tolerance of the R-tree in meters.

        Returns
        -------
        RTree
            RTree created.
        """
        rtree_created = RTree(cls.__stub.Create(messages.double_message(tolerance)))
        return rtree_created

    def _handle_rtree_obj(self, rtree_obj):
        if rtree_obj._unique_id is None:
            if (rtree_obj.obj, rtree_obj.polygon) in self._obj_to_id.keys:
                rtree_obj._unique_id = self._obj_to_id[(rtree_obj.obj, rtree_obj.polygon)]
                return True
            else:
                raise Exception("RTree object does not exist in the RTree.")
        else:
            return True

    @property
    @parser.to_box
    def extent(self) -> tuple[PointData, PointData]:
        """
        :obj:`tuple` of (:class:`.PointData`, :class:`.PointData`): Bounding box \
        for the contents of the RTree.

        This property is read-only.
        """
        return self.__stub.GetExtent(self.msg)

    def insert(self, rtree_obj: RTreeObj):
        """Insert an RTree object from a given RTree object.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with an index.
        """
        unique_id = 1 if self._unique_id is None else self._unique_id + 1
        self.__stub.InsertIntObject(RTree._r_tree_obj_message(self, rtree_obj.polygon, unique_id))
        self._unique_id = unique_id
        rtree_obj._unique_id = int(unique_id)
        self._id_to_obj[unique_id] = rtree_obj
        self._obj_to_id[(rtree_obj.obj, rtree_obj.polygon)] = unique_id

    def delete(self, rtree_obj: RTreeObj):
        """Delete the RTree from a given RTree object.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with index in this form: ``(polygon, id pair)``.
        """
        if self._handle_rtree_obj(rtree_obj):
            self.__stub.DeleteIntObject(
                RTree._r_tree_obj_message(self, rtree_obj.polygon, rtree_obj._unique_id)
            )
            del self._id_to_obj[rtree_obj._unique_id]
            del self._obj_to_id[(rtree_obj.obj, rtree_obj.polygon)]

    def empty(self) -> bool:
        """Determine if the RTree is emppty (contains no geometry).

        Returns
        -------
        bool
            ``True`` if the RTree is empty, ``False`` otherwise.
        """
        msg_empty = self.__stub.Empty(self.msg).value
        return msg_empty

    def search(self, box: tuple[PointData, PointData], bb_search: bool) -> list[RTreeObj]:
        """Search all objects intersecting a given box.

        Parameters
        ----------
        box : tuple of (.PointData, .PointData)
            Testing region, described as a "lower-left, upper-right" box.
        bb_search : bool
            Whether the RTree object intersects when the bounding-box of its
            .PolygonData instance
            intersects the testing  object. If ``False``, an explicit intersection
            is required for a hit.

        Returns
        -------
        list of .RTreeObj
           List of intersecting RTree objects.
        """
        msg = self.__stub.Search(
            pb.RTreeSearchMessage(
                target=messages.edb_obj_message(self),
                box=messages.box_message(box[0], box[1]),
                bb_search=bb_search,
            )
        )
        return [self._id_to_obj[int(to_id)] for to_id in msg.props]

    def nearest_neighbor(self, rtree_obj: RTreeObj) -> tuple[RTreeObj, tuple[PointData, PointData]]:
        """Find the nearest neighbor of a given RTree object.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree object with index in this form: ``(polygon, id pair)``.

        Returns
        -------
        tuple of (.RTreeObj, tuple of (.PointData, .PointData))

            - :class:`.RTreeObj`: Nearest-neighbor in the RTree to the provided object, or null if nothing is found.
            - tuple of (:class:`.PointData`, :class:`.PointData`) : \
                Line segment spanning the closest points between the object and the nearest neighbor.
        """
        if self._handle_rtree_obj(rtree_obj):
            msg = self.__stub.NearestNeighbor(
                RTree._r_tree_obj_message(self, rtree_obj.polygon, rtree_obj._unique_id)
            )
            return self._id_to_obj[int(msg.id)], parser.to_box(msg.coordinates)

    def touching_geometry(self, rtree_obj: RTreeObj, increment_visit: bool) -> list[RTreeObj]:
        """Find all geometries touching an RTree object.

        The provided RTree object is not returned in the touching list.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with index in this form: ``(polygon, id pair)``.
        increment_visit: bool
            Whether to increment the visit counter for items returned in the list
            of connected geometries.

        Returns
        -------
        list of .RTreeObj
            All touching RTree objects.
        """
        if self._handle_rtree_obj(rtree_obj):
            msg = self.__stub.TouchingGeometry(
                RTree._r_tree_geometry_request_message(
                    self, rtree_obj.polygon, rtree_obj._unique_id, increment_visit
                )
            )
            return [self._id_to_obj[int(to_id)] for to_id in msg.props]

    def connected_geometry(self, rtree_obj: RTreeObj, increment_visit: bool) -> list[RTreeObj]:
        """Find the connected geometries.

        If a connections exists, the provided RTree object is returned in the connected list.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with index in this form: ``(polygon, id pair)``.
        increment_visit: bool
            Whether to increment the visit counter for items returned in the connected list.

        Returns
        -------
        list of .RTreeObj
            List of connected geometries.
        """
        if self._handle_rtree_obj(rtree_obj):
            msg = self.__stub.ConnectedGeometry(
                RTree._r_tree_geometry_request_message(
                    self, rtree_obj.polygon, rtree_obj._unique_id, increment_visit
                )
            )
            return [self._id_to_obj[int(to_id)] for to_id in msg.props]

    @property
    def connected_geometry_sets(self) -> list[list[RTreeObj]]:
        """
        :obj:`list` of :obj:`list` of :class:`.RTreeObj`: Connected geometry sets of an RTree \
        in this form: ``(ids, sizes)``.

        This property is read-only.
        """
        msg = self.__stub.GetConnectedGeometrySets(messages.edb_obj_message(self))
        set_start = 0
        rtree_obj_sets = []
        for set_size in range(0, len(msg.sizes) - 1):
            rtree_obj_set = []
            for j in range(set_start, set_size):
                rtree_obj_set.append(self._id_to_obj[msg.id[j]])
            set_start += set_size
            rtree_obj_sets.append(rtree_obj_set)
        return rtree_obj_sets

    def increment_visit(self):
        """Increment the visit count, effectively marking all items in the tree as unvisited."""
        self.__stub.IncrementVisit(messages.edb_obj_message(self))

    def is_visited(self, rtree_obj: RTreeObj) -> bool:
        """Determine whether an RTree object has been visited.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with index in this form: ``(polygon, id pair)``.

        Returns
        -------
        bool
            ``True`` if the Rtree object has been visited, ``False`` otherwise.
        """
        if self._handle_rtree_obj(rtree_obj):
            return self.__stub.IsVisited(
                RTree._r_tree_obj_message(self, rtree_obj.polygon, rtree_obj._unique_id)
            ).value

    def visit(self, rtree_obj: RTreeObj):
        """Increment the count of a given RTree object.

        Parameters
        ----------
        rtree_obj : .RTreeObj
            R-tree data object with index in this form: ``(polygon, id pair)``.
        """
        if self._handle_rtree_obj(rtree_obj):
            self.__stub.Visit(
                RTree._r_tree_obj_message(self, rtree_obj.polygon, rtree_obj._unique_id)
            )

    @property
    def get_visit(self) -> int:
        """
        :obj:`int`: Visit count for the R-tree.

        This property is read-only.
        """
        return self.__stub.GetVisit(messages.edb_obj_message(self)).value

    @staticmethod
    def _r_tree_obj_message(rtree, polygon, prop_id):
        """Create an RTree object message."""
        return pb.RTreeObjMessage(
            target=messages.edb_obj_message(rtree),
            polygon=messages.polygon_data_message(polygon),
            prop=prop_id,
        )

    @staticmethod
    def _r_tree_geometry_request_message(rtree, polygon, prop_id, increment_visit):
        """Create an RTree geometry request message."""
        return pb.RTreeGeometryRequestMessage(
            target=messages.edb_obj_message(rtree),
            polygon=messages.polygon_data_message(polygon),
            prop=prop_id,
            increment_visit=increment_visit,
        )



"""Layout instance."""
import ansys.api.edb.v1.layout_instance_pb2 as layout_instance_pb2

from ansys.edb.core.geometry.point_data import PointData
from ansys.edb.core.geometry.polygon_data import PolygonData
from ansys.edb.core.inner import ObjBase, utils
from ansys.edb.core.inner.messages import (
    layer_ref_message,
    net_ref_message,
    point_message,
    polygon_data_message,
    strings_message,
)
from ansys.edb.core.layout_instance import layout_obj_instance
from ansys.edb.core.session import LayoutInstanceServiceStub, StubAccessor, StubType
from ansys.edb.core.utility.io_manager import get_cache


class LayoutInstance(ObjBase):
    """Represents a layout instance object."""

    __stub: LayoutInstanceServiceStub = StubAccessor(StubType.layout_instance)

    def refresh(self):
        """Refresh the layout instance so it contains an up-to-date geometry."""
        self.__stub.Refresh(self.msg)

    def query_layout_obj_instances(self, layer_filter=None, net_filter=None, spatial_filter=None):
        """Query layout object instances using the provided filters.

        Parameters
        ----------
        layer_filter : list[:class:`.Layer` or str or None], default: None
            Layers to query. The default is ``None``, in which case all layers are queried.
        net_filter : list[:class:`.Net` or str or None], default: None
            Nets to query. The default is ``None``, in which case all nets are queried.
        spatial_filter : :class:`.PolygonData` or
         :class:`.PointData` or ``None``, default: None
            Area of the design to query. The default is ``None``, in which case the entire
            spatial domain of the design is queried.

        Returns
        -------
        list[:class:`.LayoutObjInstance`] or tuple[list[:class:`.LayoutObjInstance`], list[:class:`.LayoutObjInstance`]]
            If a polygonal spatial filter is specified, a tuple of lists of hits is returned in this
            format: ``[<hits_completely_enclosed_in_polygon_region>, <hits_partially_enclosed_in_polygon_region>]``.
            Otherwise, a list containing all hits is returned.
        """

        def to_msg_filter_list(client_filter, ref_msg_type):
            return (
                utils.map_list(client_filter, ref_msg_type) if client_filter is not None else None
            )

        msg_params = {
            "layout_inst": self.msg,
            "layer_filter": to_msg_filter_list(layer_filter, layer_ref_message),
            "net_filter": to_msg_filter_list(net_filter, net_ref_message),
        }
        if spatial_filter is not None:
            is_point_filter = isinstance(spatial_filter, PointData)
            spatial_filter_field = "point_filter" if is_point_filter else "region_filter"
            spatial_filter_msg = (
                point_message(spatial_filter)
                if is_point_filter
                else polygon_data_message(spatial_filter)
            )
            msg_params[spatial_filter_field] = spatial_filter_msg

        request = layout_instance_pb2.LayoutObjInstancesQueryMessage(**msg_params)
        all_hits = []
        if (cache := get_cache()) is not None:
            for hits_chunk in self.__stub.StreamLayoutObjInstancesQuery(request):
                all_hits.append(hits_chunk)
        else:
            all_hits.append(self.__stub.QueryLayoutObjInstances(request))

        if isinstance(spatial_filter, PolygonData):
            full_hits = []
            partial_hits = []
            for hits in all_hits:
                full_partial_hits = hits.full_partial_hits
                for full_hit in full_partial_hits.full.items:
                    full_hits.append(layout_obj_instance.LayoutObjInstance(full_hit))
                for partial_hit in full_partial_hits.partial.items:
                    partial_hits.append(layout_obj_instance.LayoutObjInstance(partial_hit))
                return full_hits, partial_hits
        else:
            return [
                layout_obj_instance.LayoutObjInstance(hit)
                for hits in all_hits
                for hit in hits.hits.items
            ]

    def get_layout_obj_instance_in_context(self, layout_obj, context):
        """Get the layout object instance of the given :term:`connectable <Connectable>` in the provided context.

        Parameters
        ----------
        layout_obj : :term:`Connectable <Connectable>`
            Layout object with the instances to search.
        context : list[str]
            List of strings specifying the :class:`context <.LayoutInstanceContext>` that the /
            layout object instance is retrieved from.

            .. seealso:: :func:`context <ansys.edb.core.layout_instance.layout_obj_instance.LayoutObjInstance.context>`

        Returns
        -------
        :class:`.LayoutObjInstance`
        """
        return layout_obj_instance.LayoutObjInstance(
            self.__stub.GetLayoutObjInstanceInContext(
                layout_instance_pb2.GetLayoutObjInstanceInContextMessage(
                    layout_inst=self.msg,
                    layout_obj=layout_obj.msg,
                    context=strings_message(context),
                )
            )
        )

    def get_connected_objects(self, origin_layout_obj_inst, touching_only):
        """Get the layout object instance connected to the origin layout object instance.

        Parameters
        ----------
        origin_layout_obj_inst : :class:`.LayoutObjInstance`
            Layout object instance that is to act as the origin to get connected objects from.
        touching_only : bool
            Whether to get only layout object instances touching the orgigin layout object instance
            on the placement layer of the origin layout object. If ``False`` all layout object
            instances across all layers that are electrically connected to the origin layout object
            instance are returned.

        Returns
        -------
        list[:class:`.LayoutObjInstance`]
           List of layout object instances connected to the origin layout object instance.
        """
        hits = self.__stub.GetConnectedObjects(
            layout_instance_pb2.GetConnectedObjectsMessage(
                layout_inst=self.msg,
                layout_obj=origin_layout_obj_inst.msg,
                touching_only=touching_only,
            )
        )
        return utils.map_list(hits.items, layout_obj_instance.LayoutObjInstance)



"""Dielectric material definition."""
from ansys.api.edb.v1 import multipole_debye_model_pb2_grpc
import ansys.api.edb.v1.multipole_debye_model_pb2 as pb
from google.protobuf import empty_pb2

from ansys.edb.core import session
from ansys.edb.core.definition.dielectric_material_model import DielectricMaterialModel
from ansys.edb.core.inner import messages


class MultipoleDebyeModel(DielectricMaterialModel):
    """Represents a dielectric material model."""

    __stub: multipole_debye_model_pb2_grpc.MultipoleDebyeModelServiceStub = session.StubAccessor(
        session.StubType.multipole_debye_model
    )

    @classmethod
    def create(cls):
        """Create a multipole Debye dielectric material model.

        Returns
        -------
        MultipoleDebyeModel
        """
        return MultipoleDebyeModel(cls.__stub.Create(empty_pb2.Empty()))

    def get_parameters(self):
        """Get parameters used to define the model.

        Returns
        -------
        frequencies : list[float]
            List of frequencies.
        permitivities : list[float]
            List of relative permitivites at each frequency.
        loss_tangents : list[float]
            List of loss tangents at each frequency.
        """
        parameters_msg = self.__stub.GetParameters(messages.edb_obj_message(self))
        return (
            [i.value for i in parameters_msg.frequencies],
            [i.value for i in parameters_msg.relative_permitivities],
            [i.value for i in parameters_msg.loss_tangents],
        )

    def set_parameters(self, frequencies, permitivities, loss_tangents):
        """Set parameters used to define the model.

        Parameters
        ----------
        frequencies : list[float]
            List of frequencies.
        permitivities : list[float]
            List of relative permitivites at each frequency.
        loss_tangents : list[float]
            List of loss tangents at each frequency.
        """
        frequencies_msg = [messages.double_message(i) for i in frequencies]
        permitivities_msg = [messages.double_message(i) for i in permitivities]
        loss_tangents_msg = [messages.double_message(i) for i in loss_tangents]
        self.__stub.SetParameters(
            pb.MultipoleDebyeModelSetParams(
                target=self.msg,
                vectors=pb.MultipoleDebyeModelGetParams(
                    frequencies=frequencies_msg,
                    relative_permitivities=permitivities_msg,
                    loss_tangents=loss_tangents_msg,
                ),
            )
        )



"""SPICE Model."""
from ansys.edb.core.hierarchy.model import Model
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.session import SpiceModelServiceStub, StubAccessor, StubType


class SPICEModel(Model):
    """Represents a SPICE model object."""

    __stub: SpiceModelServiceStub = StubAccessor(StubType.spice_model)

    @classmethod
    def create(cls, name, path, sub_circuit):
        """Create a SPICE model.

        Parameters
        ----------
        name : str
            Name of the SPICE model file.
        path : str
            Path to the SPICE model file.
        sub_circuit : str
            Subcircuit name of the SPICE model.
        """
        return cls(cls.__stub.Create(messages.spice_model_message(name, path, sub_circuit)))

    @property
    def _properties(self):
        return self.__stub.GetProperties(messages.edb_obj_message(self))

    @property
    def model_name(self):
        """:obj:`str`: Name of the SPICE model file."""
        return self._properties.name

    @model_name.setter
    def model_name(self, name):
        self.__stub.SetModelName(messages.string_property_message(self, name))

    @property
    def model_path(self):
        """:obj:`str`: Path to the SPICE model file."""
        return self._properties.path

    @model_path.setter
    def model_path(self, path):
        self.__stub.SetModelPath(messages.string_property_message(self, path))

    @property
    def sub_circuit(self):
        """:obj:`str`: Name of the subcircuit in the SPICE model."""
        return self._properties.sub_ckt

    @sub_circuit.setter
    def sub_circuit(self, name):
        self.__stub.SetSubCkt(messages.string_property_message(self, name))

    def add_terminal(self, terminal, pin):
        """Add a terminal with a pin number.

        Parameters
        ----------
        terminal : str
            Terminal name to associate with the pin.
        pin : str
            Pin number.
        """
        self.__stub.AddTerminalPinPair(
            messages.spice_model_net_terminal_pin_message(self, terminal, pin)
        )

    def remove_terminal(self, terminal):
        """Remove a terminal.

        Parameters
        ----------
        terminal : str
            Terminal name.
        """
        self.__stub.RemoveTerminalPinPair(messages.string_property_message(self, terminal))

    @property
    @parser.to_string_dict
    def terminal_pin_pairs(self):
        """:obj:`dict` of {:obj:`str` : :obj:`str`}: A dictionary showing the mapping of terminals in the layout to \
        pins in the spice model by name."""
        return self.__stub.GetTerminalPinPairs(self.msg)



"""Module containing shared decorators."""
import functools


def _as_array(array_or_item):
    try:
        iter(array_or_item)
        return array_or_item
    except TypeError:
        return [array_or_item]


def _self_as_param_class_method(fn):
    @functools.wraps(fn)
    def wrapper(first, second):
        return fn(_as_array(first), _as_array(second))

    return wrapper


def _self_as_param_instance_method(fn, instance):
    @functools.wraps(fn)
    def wrapper(other):
        return fn(_as_array(instance), _as_array(other))

    return wrapper


class self_as_param(classmethod):  # noqa
    def __init__(self, func):
        """Initialize a decorator."""
        super().__init__(func)

    def __get__(self, instance, owner):
        """Fetch decorated function."""
        fn = super().__get__(instance, owner)
        if instance is None:
            return _self_as_param_class_method(fn)
        else:
            return _self_as_param_instance_method(fn, instance)



"""Cell instance."""

from ansys.api.edb.v1.cell_instance_pb2_grpc import CellInstanceServiceStub

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.hierarchy import hierarchy_obj
from ansys.edb.core.inner import messages
from ansys.edb.core.layout import layout
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.transform3d import Transform3D
from ansys.edb.core.utility.value import Value


class CellInstance(hierarchy_obj.HierarchyObj):
    """Represents a cell instance object."""

    __stub: CellInstanceServiceStub = StubAccessor(StubType.cell_instance)
    layout_obj_type = LayoutObjType.CELL_INSTANCE

    @classmethod
    def create(cls, layout, name, ref):
        """Create a cell instance with a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the cell instance in.
        name : str
            Name of the cell instance.
        ref : :class:`.Layout`
            Layout that the cell instance refers to.

        Returns
        -------
        CellInstance
            Cell instance created.
        """
        return CellInstance(
            cls.__stub.Create(messages.cell_instance_creation_message(layout, name, ref))
        )

    @classmethod
    def create_with_component(cls, layout, name, ref):
        """Create a cell instance with a component definition.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the cell instance in.
        name : str
            Name of the cell instance.
        ref : :class:`.ComponentDef`
            Component that the cell instance refers to.

        Returns
        -------
        CellInstance
            Cell instance created.
        """
        return CellInstance(
            cls.__stub.CreateWithComponent(
                messages.cell_instance_creation_message(layout, name, ref)
            )
        )

    @classmethod
    def find(cls, layout, name):
        """Find a cell instance by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the cell instance.
        name : str
            Name of the cell instance.

        Returns
        -------
        CellInstance
            Cell instance that is found, ``None`` otherwise.
        """
        return CellInstance(
            cls.__stub.FindByName(messages.object_name_in_layout_message(layout, name))
        )

    @property
    def reference_layout(self):
        """:class:`.Layout`: Reference layout of the cell instance.

        This property is read-only.
        """
        return layout.Layout(self.__stub.GetReferenceLayout(self.msg))

    @property
    def term_instances(self):
        """:obj:`list` of :class:`.TerminalInstance`: Terminal \
        instances associated with the cell instance.

        This property is read-only.
        """
        from ansys.edb.core.terminal.terminal_instance import TerminalInstance

        terms = self.__stub.GetTermInsts(self.msg).items
        return [TerminalInstance(ti) for ti in terms]

    @property
    def placement_3d(self):
        """:obj:`bool`: Flag indicating if the cell instance is 3D placed in the owning layout.

        If ``True``, transformation can be set using the :obj:`transform3d` object.
        """
        return self.__stub.GetIs3DPlacement(self.msg).value

    @placement_3d.setter
    def placement_3d(self, value):
        """Set the cell instance as 3D placed in the owning layout."""
        self.__stub.Set3DPlacement(messages.bool_property_message(self, value))

    @property
    def transform3d(self):
        """:class:`.Transform3D`: \
        3D transformation information of the cell instance.

        For the transformation to be applied, the :obj:`placement_3d` property
        must be set to ``True``.
        """
        return Transform3D(self.__stub.Get3DTransform(self.msg))

    @transform3d.setter
    def transform3d(self, value):
        """Set the 3D transformation for the cell instance.

        The cell instance must be 3D placed.
        """
        self.__stub.Set3DTransform(messages.pointer_property_message(self, value))

    def get_parameter_override(self, param_name):
        """Get the override of the cell instance parameter by name.

        Parameters
        ----------
        param_name : str
            Name of the cell instance parameter.

        Returns
        -------
        :class:`.Value`
            Override value for the cell instance parameter.
        """
        return Value(
            self.__stub.GetParameterOverride(messages.string_property_message(self, param_name))
        )

    def set_parameter_override(self, param_name, param_value):
        """Set an override value for a given cell instance parameter.

        Parameters
        ----------
        param_name : str
            Name of the cell instance parameter.
        param_value : :class:`.Value`
            Value to override with.
        """
        self.__stub.SetParameterOverride(
            messages.cell_instance_parameter_override_message(self, param_name, param_value)
        )
